""" This file defines the class for the TensorBoard logger. """
import importlib.util
import os
from typing import Any, Optional, Union

from .base_logger import Logger
from .logger_utils import NotImportedModule, value_to_float

try:  # Tensorboard integration, see original package : https://www.tensorflow.org/tensorboard
    import tensorflow
except ImportError:
    tensorflow = NotImportedModule("tensorflow")

class TensorBoardLogger(Logger):
    """ TensorBoard Logger. This logger logs things to a logging directory (logdir). """
    def __init__(self, tracker):
        super().__init__("TensorBoard Logger", tracker)
        self.writers = None
        self.logdir = self.tracker.config.get("logdir", None)

    def check_config_requirements(self) -> str:
        if self.logdir is None:
            return ("Config Requirement Error : no logdir was provided for the tensorboard logger. Provide one using "
                    "the logdir key in the tracker config.")
        return ""

    def check_install(self) -> str:
        if not importlib.util.find_spec("tensorflow"):
            return "Install Error : your experiment tracking config requires tensorflow - currently not installed !"
        return ""

    def get_logger_object(self) -> Any:
        return self.writers

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        if "%e" not in self.logdir:
            writer_path = os.path.join(self.logdir, experiment_name, run_name)
        else:
            writer_path = self.logdir.replace("%e", self.tracker.experiment.config.get_experiment_path())
        self.writers = {
            sub_logger: tensorflow.summary.create_file_writer(os.path.join(writer_path, sub_logger))
            for sub_logger in self.tracker.sub_loggers}

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        value = value_to_float(value, self.name)
        if value == "":
            return

        tb_sub_logger = "" if sub_logger is None else sub_logger
        if tb_sub_logger not in self.writers:
            raise ValueError(f"Sub-logger '{tb_sub_logger}' is not defined in the tracker config. You can only use "
                             "tensorboard loggers that have been defined using the 'sub_loggers' key in the tracker "
                             "config.")
        with self.writers[tb_sub_logger].as_default():
            tensorflow.summary.scalar(name=name, data=value, step=0 if step is None else step, description=description)



"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import os
from contextlib import ExitStack
from functools import partial
from typing import Any, Callable, List, Optional

from .experiment_utils import format_mode
from .tracker import Tracker
from ..config.config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class Experiment:
    """ Class automating tracking using different tracking packages. """

    def __init__(self, config: Configuration, main_function: Callable,
                 experiment_name: Optional[str] = None, run_name: Optional[str] = None,
                 params_filter_fn: Optional[Callable[[Configuration], List[str]]] = None,
                 log_modified_params_only: bool = False, do_not_log_hooks: bool = False,
                 only_params_to_log: Optional[List[str]] = None, params_not_to_log: Optional[List[str]] = None,
                 description_formatter: Optional[Callable[[Optional[str]], str]] = None):
        """
        Creates an instance of the Experiment class, which wraps around a main function.

        :param config: config used for the experiment
        :param main_function: function to run to perform the experiment
        :param experiment_name: name of the experiment, defaults to name
            of the folder set as experiment path in the config
        :param run_name: name of the run, defaults as the index of the run in the experiment folder
        :param params_filter_fn: function to use instead of the default filter to get the list of the names of the
            parameters to log to the tracker from the config. If this is used, then 'log_modified_params_only',
            'only_params_to_log' and 'params_not_to_log' are ignored.
        :param log_modified_params_only: whether the parameters to filter using the other arguments are the parameters
            that changed compared to the default config (True) or only those of the whole config (False)
        :param do_not_log_hooks: whether the parameters defined as hooks should be filtered out (True) or not (False)
        :param only_params_to_log: if provided, only the parameters whose names are given will be filtered and logged
        :param params_not_to_log: if provided, parameters whose names are given will be filtered out
        :param description_formatter: optional function to use to format the provided run description instead of the
            default formatter self.default_formatter
        """
        self.config = config
        self.main_function = main_function
        if not hasattr(self, "number_of_runs"):
            self.number_of_runs = None
        self.current_run = None
        self.format_description = self.default_formatter if description_formatter is None else description_formatter
        tracker_config = self.config.get_hook("tracker_config")
        if len(tracker_config) > 1:
            raise RuntimeError("Several parameters were registered as tracker configs. Please register only one.")
        self.tracker = Tracker(self.config[tracker_config[0]] if tracker_config else {"type": ""},
                               self, experiment_name=experiment_name, run_name=run_name, starting_step=0,
                               params_filter_fn=params_filter_fn, log_modified_params_only=log_modified_params_only,
                               do_not_log_hooks=do_not_log_hooks, only_params_to_log=only_params_to_log,
                               params_not_to_log=params_not_to_log)

    def default_formatter(self, description: Optional[str]) -> str:
        """
        This function formats the provided description before passing it to the trackers.

        :param description: provided description to format. You can use flags in the description, which will be used
            during formatting.
            - `%h` (can be used once in the description) : Everything before this tag will be considered the header of
              the description
            - `%m` : Will be replaced by the mode of the experiment (train, test, etc.)
            - `%n` : Will be replaced by the name of the experiment
            - `%p` : Will be replaced by the path of the experiment
            - '%v' : Will be replaced by the variation name of the experiment
        :raises RuntimeError: when more than one header tag is detected in the description
        :return: formatted description
        """
        # Setup
        try:
            experiment_path = self.config.get_experiment_path()
        except RuntimeError:
            experiment_path = None
        mode_param = self.config.get_hook("mode")
        mode = format_mode(self.config[mode_param[0]]) if mode_param else "unspecified mode"
        description = description if description else "unstated purpose"
        if "%n" in description:
            name = None if experiment_path is None else os.path.basename(experiment_path)
            description = description.replace("%n", str(name))
        if "%p" in description:
            description = description.replace("%p", str(experiment_path))
        if "%m" in description:
            description = description.replace("%m", str(mode))
        if "%v" in description:
            description = description.replace("%v", str(self.config.get_variation_name()))

        # Header
        header = "Run" + ("" if experiment_path is None else f" in {experiment_path}")
        if mode != "unspecified mode":
            header = f"{mode} r{header[1:]}"
        if "%h" in description:
            if description.count("%h") > 1:
                raise RuntimeError("You can only declare one header in the run description.")
            header, description = description.split('%h')
        header = header[0].capitalize() + header[1:]

        # Final description
        description = f"{header}\nPurpose : {description}"
        if self.number_of_runs > 1:
            description = f"[{self.current_run + 1}/{self.number_of_runs}] {description}"
        return description

    def run(self, run_description: Optional[str] = None, **kwargs) -> Any:
        """
        Creates all variations of the config and starts a run for each of them.

        :param run_description: if passed, will serve as a description for the purpose of the current run. This
            description will be formatted using either the default formatter or a formatter passed when creating the
            Experiment object. With the default formatter, you can use flags in the description, which will be used
            during formatting.
            - `%h` (can be used once in the description) : Everything before this tag will be considered the header of
              the description
            - `%m` : Will be replaced by the mode of the experiment (train, test, etc.)
            - `%n` : Will be replaced by the name of the experiment
            - `%p` : Will be replaced by the path of the experiment
            - '%v' : Will be replaced by the variation name of the experiment
        :param kwargs: arguments to pass to the main function aside from the config and the tracker
        :return: whatever the main function returns
        """
        variations = self.config.create_variations()
        self.number_of_runs = len(variations)
        description = run_description
        returns = []
        for run_number, variation in enumerate(variations):
            self.current_run = 0
            variation_description = description
            if variation is not self.config:
                filter_fn = (None if self.tracker.get_filtered_params is self.tracker.default_filter
                             else self.tracker.get_filtered_params)
                format_fn = None if self.format_description is self.default_formatter else self.format_description
                self.__init__(config=variation,  # pylint: disable=unnecessary-dunder-call
                              main_function=self.main_function, experiment_name=self.tracker.experiment_name,
                              run_name=self.tracker.run_name, params_filter_fn=filter_fn,
                              log_modified_params_only=self.tracker.log_modified_params_only,
                              only_params_to_log=self.tracker.only_params if filter_fn is None else None,
                              params_not_to_log=self.tracker.except_params if filter_fn is None else None,
                              description_formatter=format_fn)
                self.current_run = run_number
                description = (input("[TRACKER] Please describe the purpose of these runs : ")
                               if description is None else description)
                description = description if description else "unstated purpose"
                if description.count("%h") == 1:
                    variation_description = (f"{description.split('%h')[0]} (variation %v)%h"
                                             f"{description.split('%h')[1]}")
                else:
                    variation_description = f"%m run in %p (variation %v) :%h{description}"
            try:
                to_return = self.run_single(variation_description, **kwargs)
            except Exception as exception:  # pylint: disable=broad-except
                YAECS_LOGGER.error(f"Run {run_number + 1} failed with the following error : {exception}")
                to_return = None
            returns.append(to_return)
        return returns

    def run_single(self, run_description: Optional[str] = None, **kwargs) -> Any:
        """
        Runs a single experiment with the defined config. Does not check the config variations.

        :param run_description: if passed, will serve as a description for the purpose of the current run. This
            description will be formatted using either the default formatter or a formatter passed when creating the
            Experiment object. With the default formatter, you can use flags in the description, which will be used
            during formatting.
            - `%h` (can be used once in the description) : Everything before this tag will be considered the header of
              the description
            - `%m` : Will be replaced by the mode of the experiment (train, test, etc.)
            - `%n` : Will be replaced by the name of the experiment
            - `%p` : Will be replaced by the path of the experiment
            - '%v' : Will be replaced by the variation name of the experiment
        :param kwargs: arguments to pass to the main function aside from the config and the tracker
        :return: whatever the main function returns
        """
        # Start the run
        if os.getenv('NODE_RANK'):  # do not track if in a pytorch-lightning spawned process
            description = None
        else:
            description = (input("[TRACKER] Please describe the purpose of this run : ")
                           if run_description is None else run_description)
            description = self.format_description(description)
        self.tracker.start_run(description=description)

        # Function and context preparation
        main_function = partial(self.main_function, config=self.config, tracker=self.tracker, **kwargs)
        main_function.__name__ = self.main_function.__name__  # partial functions do not have names
        main_function = self.tracker.loggers.modify_main_function(main_function)
        contexts = self.tracker.loggers.main_function_context()

        # Run
        with ExitStack() as stack:
            for context in contexts:
                stack.enter_context(context)
            to_return = main_function()
        return to_return



""" This file defines the class for the Sacred logger. """
from functools import partial
import importlib.util
import logging
from typing import Any, Optional, Union

from .base_logger import Logger
from .logger_utils import NotImportedModule, value_to_float

try:  # Sacred integration, see original package : https://sacred.readthedocs.io/
    import sacred
except ImportError:
    sacred = NotImportedModule("sacred")

YAECS_LOGGER = logging.getLogger(__name__)


class SacredLogger(Logger):
    """ Sacred Logger. This logger logs things to a setup database. """
    def __init__(self, tracker):
        super().__init__("Sacred Logger", tracker)
        self.experiment = None
        self.db_name = self.tracker.config.get("db_name", None)
        self.db_url = self.tracker.config.get("db_url", None)
        self.description: Optional[str] = None

    def check_config_requirements(self) -> str:
        errors = ""
        if self.db_name is None:
            errors += ("Config Requirement Error : no db_name was provided for the sacred logger. Provide one "
                       "using the db_name key in the tracker config.")
        if self.db_url is None:
            errors += "\n" if errors else ""
            errors += ("Config Requirement Error : no db_url was provided for the sacred logger. Provide one "
                       "using the db_url key in the tracker config.")
        return errors

    def check_install(self) -> str:
        if not importlib.util.find_spec("sacred"):
            return "Install Error : your experiment tracking config requires sacred - currently not installed !"
        return ""

    def modify_main_function(self, main_function):
        self.experiment.main(main_function)
        main_function = partial(self.experiment.run, meta_info={"comment": self.description})
        main_function.__name__ = self.tracker.experiment.main_function.__name__  # partial functions do not have names
        return main_function

    def get_logger_object(self) -> Any:
        return self.experiment

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        self.experiment = sacred.Experiment(f"{experiment_name}/{run_name}")
        self.experiment.observers.append(sacred.observers.MongoObserver(url=self.db_url, db_name=self.db_name))
        self.experiment.add_config(params)
        self.description = description

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        self._warn_function_argument("log_scalar", "description", description, None)
        value = value_to_float(value, self.name)
        if value == "":
            return

        extended_name = name
        if sub_logger is not None:
            extended_name = f"{sub_logger}/{extended_name}"
        self.experiment.log_scalar(name=extended_name, value=value, step=step)



""" This module contains classes to track and manage time durations. """

import logging
import time
from typing import Dict, List, Optional, Tuple, Union

from ..yaecs_utils import compare_string_pattern

YAECS_LOGGER = logging.getLogger(__name__)


class WildCardDict(dict):
    """ This class is a dict where getitem also supports None, lists of items and wildcards ('*'). """
    def __getitem__(self, item: Union[None, str, List[str]]):
        if item is None:
            return self
        if isinstance(item, str) and "*" not in item:
            return super().__getitem__(item)
        if isinstance(item, str):
            return {key: super().__getitem__(key) for key in self.keys() if compare_string_pattern(key, item)}
        return [self.__getitem__(pattern) for pattern in item]


class Timer:
    """ This class is a timer. It has a name to know what it records, and start and stop times. """
    def __init__(self, name: str = "MyTimer", verbose: Optional[int] = 1, start: bool = False,
                 step: Optional[int] = None, start_time: Optional[float] = None):
        """
        Creates a timer.

        :param name: name of the timer
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :param start: whether to start the timer immediately
        :param step: starting step if start is True (if None, assumes step=0, timings are averages over elapsed steps)
        :param start_time: optional time to start the timer at if start is True (if None, uses current time)
        """
        current_time = time.time() if start_time is None else start_time
        if "*" in name:
            raise ValueError(f"Invalid name for Timer : '{name}'."
                             "The special character '*' cannot be used in Timer names.")
        self.name: str = name
        self.start_times: List[Tuple[Union[float, int]]] = []
        self.stop_times: List[Tuple[Union[float, int]]] = []
        self.verbose: int = 1 if verbose is None else verbose
        if start:
            self.start(step=step, start_time=current_time)

    def __str__(self) -> str:
        return self.render("current") if self.running else self.render("last")

    def __getitem__(self, item: Union[int, str]) -> Optional[float]:
        return self.get(which=item, step_aggregation="average")

    @property
    def empty(self) -> bool:
        """ Returns whether the timer has completed any run. """
        return len(self.stop_times) == 0

    @property
    def running(self) -> bool:
        """ Returns whether the timer is currently running. """
        return len(self.start_times) > len(self.stop_times)

    @property
    def steps(self) -> List[int]:
        """ Returns the steps on which the timer was recorded. """
        steps = set()
        for i in self.timings:
            for j in range(self.start_times[i][1], self.stop_times[i][1] + 1):
                steps.add(j)
        return sorted(steps)

    @property
    def timings(self) -> List[int]:
        """ Returns the timings recorded by the timer. """
        return list(range(len(self.stop_times)))

    @property
    def verbose(self) -> int:
        """ Returns the verbosity level of the timer. """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose: int) -> None:
        """ Sets the verbosity level of the timer. """
        if verbose not in [0, 1, 2]:
            raise ValueError(f"Invalid value for 'verbose' : {verbose}. Can be 0, 1 or 2.")
        self._verbose = verbose

    def get(self, which: Union[int, str] = "last", step_aggregation: str = "average",
            current_step: Optional[int] = None) -> Optional[float]:
        """
        Gets a specified recorded timing (the last one by default). A recorded timing needs a start and a stop.

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'last' (last recorded timing), 'first' (first recorded timing), 'average' (average of all recorded timings),
            'total' (sum of all recorded timings) or 'current' (current timing)
        :param step_aggregation: whether to return each timing as an average over the steps it was recorded on or as a
            total. Can be 'average' or 'total'
        :param current_step: current step used to compute the current timing if which is 'current'
        :raises ValueError: if which is not an int or in ['last', 'first', 'average', 'total']
        :raises IndexError: if which is an int and is out of range
        :return: the timing
        """
        current_time = time.time()

        def _aggregate(value, number_of_steps):
            return value / number_of_steps if step_aggregation == "average" else value
        if step_aggregation not in ["average", "total"]:
            raise ValueError(f"Invalid value for 'step_aggregation' : {step_aggregation}. Can be 'average' or 'total'.")
        which = self._process_which(which, possible_strings=["last", "first", "average", "total", "current"])

        # When which is "current", return the current timing as though the timer was stopped at current_time
        if which == "current":
            if self.running:
                return _aggregate(current_time - self.start_times[-1][0],
                                  self.get_number_of_steps("current", current_step=current_step))
            return None

        if self.empty:
            return 0 if which == "total" else None

        # When which is "average", behave differently depending on step_aggregation
        if which == "average":
            if step_aggregation == "average":
                # If step_aggregation is "average", return the average of recorded steps
                return sum(self.get_at_step(i) for i in self.steps)/len(self.steps)
            # If step_aggregation is "total", return the average of recorded timing
            return sum(self.get(which=i, step_aggregation="total") for i in self.timings)/len(self.timings)

        # When which is "total", always return the total recorded duration regardless of step_aggregation
        if which == "total":
            return sum(self.get(which=i, step_aggregation="total") for i in self.timings)

        # When which is an int, return the timing at the given index
        return _aggregate(self.stop_times[which][0] - self.start_times[which][0], self.get_number_of_steps(which))

    def get_at_step(self, step: Optional[Union[int, str]] = None) -> Optional[float]:
        """
        Gets the duration of given step, or None if given step was not recorded. Similar to 'get' except it takes steps
        instead of timing indices. Also accepts 'last' and 'first' as steps.

        :param step: if None last step, else step to get
        :return: the duration if it was recorded, otherwise None
        """
        if step == "first":
            step = self.start_times[0][1]
        if step == "last":
            step = self.stop_times[-1][1]
        if step not in self.steps:
            return None
        step_duration = 0
        for timing in self.timings:
            if self.start_times[timing][1] <= step <= self.stop_times[timing][1]:
                step_duration += self.get(which=timing, step_aggregation="average")
        return step_duration

    def get_number_of_steps(self, which: Union[int, str] = "last", current_step: Optional[int] = None) -> Optional[int]:
        """
        Gets the number of steps of a specified recorded timing (the last one by default).

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'last' (last recorded timing), 'first' (first recorded timing), 'average' (average of all recorded timings),
            'total' (sum of all recorded timings) or 'current' (current timing)
        :param current_step: current step used to compute the current timing if which is 'current'
        :return: the number of steps
        """
        which = self._process_which(which, possible_strings=["last", "first", "average", "total", "current"])
        if which == "current":
            if not self.running:
                return None
            if current_step is None:
                return 1
            return 1 + current_step - self.start_times[-1][1]
        if self.empty:
            return 0 if which == "total" else None
        if which in ["average", "total"]:
            return self.stop_times[-1][1] - self.start_times[0][1]
        return 1 + self.stop_times[which][1] - self.start_times[which][1]

    def render(self, which: Union[int, str] = "current", step_aggregation: str = "average",
               current_step: Optional[int] = None, verbose: Optional[int] = None) -> str:
        """
        Renders a string to display all or part of the timer's internal state.

        :param which: which timing to get. Can be an int (index of the timing in the list of recorded timings), or
            'current' (current timing), 'last' (last recorded timing), 'first' (first recorded timing), 'average'
            (average of all recorded timings), 'total' (sum of all recorded timings) or 'all' (all recorded timings)
        :param step_aggregation: whether to return each timing as an average over the steps it was recorded on or as a
            total. Can be 'average' or 'total'
        :param current_step: current step used to compute the current timing if which is 'current'
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :raises ValueError: if verbose is not in [0, 1, 2]
        :return: the rendered string
        """
        def _render_duration(which, step_aggregation, current_step, verbose):
            rendered = ""
            duration = self.get(which=which, step_aggregation=step_aggregation, current_step=current_step)
            if duration is None:
                if verbose == 2:
                    rendered += "no timing recorded so far"
                if verbose == 1:
                    rendered += "none"
                if verbose == 0:
                    rendered += "-"
            else:
                duration = format_duration(duration, human_readable=bool(verbose))
                number_of_steps = self.get_number_of_steps(which)
                if verbose == 2:
                    timing_name = f"{which} timing" if isinstance(which, str) else f"timing {which}"
                    step_operation = "averaged" if step_aggregation == "average" else "accumulated"
                    step_rendering = f" ({step_operation} over {number_of_steps} steps)" if number_of_steps > 1 else ""
                    rendered += f"{timing_name} : {duration}{step_rendering}"
                if verbose == 1:
                    step_operation = "/" if step_aggregation == "average" else "in"
                    step_rendering = f" ({step_operation} {number_of_steps} steps)" if number_of_steps > 1 else ""
                    rendered += f"{duration}{step_rendering}"
                if verbose == 0:
                    rendered += f"{duration}"
            return rendered
        current_time = time.time()
        verbose = self.verbose if verbose is None else verbose
        if verbose not in [0, 1, 2]:
            raise ValueError(f"Invalid value for 'verbose' : {verbose}. Can be 0, 1 or 2.")
        which_tmp = self._process_which(which, possible_strings=["current", "last", "first", "average", "total", "all"])
        which = which if which in ["last", "first"] else which_tmp
        rendered_string = ""

        # Set header
        if verbose == 2:
            rendered_string += f"---- '{self.name}' timer ----\n"
            header_size = len(rendered_string)
        if verbose == 1:
            rendered_string += f"'{self.name}' : "
        if verbose == 0:
            rendered_string += f"{self.name}("

        # Set summary
        if verbose == 2:
            steps = sum(self.stop_times[i][1] - self.start_times[i][1] for i in range(len(self.stop_times)))
            rendered_string += f"Overall {len(self.stop_times)} recorded timings totalling {steps} steps.\n"

        # Set body
        if which == "current":
            if self.running:
                duration = format_duration(current_time - self.start_times[-1][0], human_readable=bool(verbose))
                if verbose == 2:
                    rendered_string += f"Currently running for {duration}.\n"
                if verbose == 1:
                    rendered_string += f"running for {duration}"
            else:
                duration = "-"
                if verbose == 2:
                    rendered_string += "Currently not running.\n"
                if verbose == 1:
                    rendered_string += "not running"
            if verbose == 0:
                rendered_string += f"{duration}"
        elif which != "all":
            duration = _render_duration(which, step_aggregation, current_step, verbose)
            if verbose == 2:
                rendered_string += f"{duration.capitalize()}.\n"
            else:
                rendered_string += duration
        else:
            if self.empty:
                if verbose == 2:
                    rendered_string += "No timing recorded so far.\n"
                if verbose == 1:
                    rendered_string += "none"
                if verbose == 0:
                    rendered_string += "-"
            else:
                if verbose == 2:
                    rendered_string += "All recorded timings :\n - "
                    rendered_string += " ;\n - ".join([_render_duration(i, step_aggregation, current_step, verbose)
                                                       for i in range(len(self.stop_times))])
                    rendered_string += ".\n"
                else:
                    rendered_string += " ; ".join([_render_duration(i, step_aggregation, current_step, verbose)
                                                  for i in range(len(self.stop_times))])

        # Set footer
        if verbose == 2:
            rendered_string += "-" * header_size
        if verbose == 1:
            rendered_string += "."
        if verbose == 0:
            rendered_string += ")"

        return rendered_string

    def reset(self) -> None:
        """ Resets the timer, removing all recorded timings. """
        self.start_times = []
        self.stop_times = []

    def start(self, step: Optional[int] = None, start_time: Optional[float] = None) -> float:
        """
        Automatically stops any previously started timer, and starts a new one.

        :param step: starting step (if none, assumes step=last stop step, timings are averages over the elapsed steps)
        :param start_time: optional time to start the timer at (if None, uses current time)
        :raises ValueError: if step is less than the previous stopping step
        :return: the starting timestamp
        """
        current_time = time.time() if start_time is None else start_time
        if self.stop_times and current_time < self.stop_times[-1][0]:
            raise ValueError(f"Invalid value for 'time' : {current_time}. Must be later than the last stopping time.")
        self.stop(step=step, stop_time=current_time)
        previous_step = 0
        if not self.empty:
            previous_step = self.stop_times[-1][1]
        step = previous_step if step is None else step
        if step < previous_step:
            raise ValueError(f"Invalid value for 'step' : {step}. Must be greater than or equal to the previous "
                             "stopping step.")
        self.start_times.append((current_time, step))
        if self.verbose == 2:
            YAECS_LOGGER.info(f"Timer '{self.name}' started at step {step} at time "
                              f"{time.asctime(time.localtime(current_time))}.")
        return current_time

    def stop(self, step: Optional[int] = None, stop_time: Optional[float] = None) -> Optional[float]:
        """
        Stops any previously started timer, or does nothing if no timer is running.

        :param step: starting step (if none assumes step=starting step + 1, timings are averages over the elapsed steps)
        :param stop_time: optional time to stop the timer at (if None, uses current time)
        :raises ValueError: if step is less than or equal to the previous starting step
        :return: the duration of the timer
        """
        current_time = time.time() if stop_time is None else stop_time
        if not self.running:
            return None
        if current_time < self.start_times[-1][0]:
            raise ValueError(f"Invalid value for 'time' : {current_time}. Must be later than the starting time.")
        previous_step = self.start_times[-1][1]
        step = previous_step if step is None else step
        if step < previous_step:
            raise ValueError(f"Invalid value for 'step' : {step}. Must be greater than the previous starting step.")
        self.stop_times.append((current_time, step))
        ellapsed = current_time - self.start_times[-1][0]
        if self.verbose == 2:
            YAECS_LOGGER.info(f"Timer '{self.name}' stopped at step {step} : ellapsed time "
                              f"{format_duration(ellapsed)} (on average "
                              f"{format_duration(ellapsed/self.get_number_of_steps('last'))} per step).")
        return ellapsed

    def _process_which(self, which: Union[int, str], possible_strings: List[str]) -> Union[int, str]:
        """
        Processes 'which' arguments.

        :raises ValueError: if which is not an int or in possible_strings
        :raises IndexError: if which is an int and is out of range
        :return: the processed which
        """
        if isinstance(which, int) and which >= len(self.stop_times):
            raise IndexError(f"Index {which} is out of range for timer {self.name} with {len(self.stop_times)} "
                             "recorded timings.")
        if not isinstance(which, int) and which not in possible_strings:
            raise ValueError(f"Invalid value for 'which' : {which}. Can be an int or in {possible_strings}.")
        if which == "last":
            which = self.timings[-1]
        elif which == "first":
            which = self.timings[0]
        return which


class TimerManager:
    """ This class records timings and prints or returns them in a variety of formats. """
    def __init__(self, verbose: Optional[int] = None):
        """
        Creates a timer manager use to start and stop timers, as well as render them.

        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail. If None, set based on local logger
        """
        self.timers: Dict[str, Timer] = {}
        level = YAECS_LOGGER.getEffectiveLevel()
        level_verbose = 0 if level > 30 else 1 if level > 10 else 2
        verbose = level_verbose if verbose is None else verbose
        self.verbose: int = verbose

    def __getitem__(self, item: Union[int, str]) -> WildCardDict:
        item = self._process_which(item)
        if item == "current":
            return WildCardDict({name: timer.get("current") for name, timer in self.timers.items()})
        if item == "average":
            return WildCardDict({name: timer.get("average", step_aggregation="average")
                                 for name, timer in self.timers.items()})
        if item == "total":
            return WildCardDict({name: timer.get("total") for name, timer in self.timers.items()})
        return WildCardDict({name: timer.get_at_step(item) for name, timer in self.timers.items()})

    @property
    def first_step(self) -> int:
        """ Gets the first step of the timer manager (assuming it is the first step of all its timers). """
        if not self.steps:
            return 0
        return min(self.steps)

    @property
    def last_step(self) -> int:
        """ Gets the last or current step of the timer manager (assuming it is the latest step of all its timers). """
        if not self.steps:
            return 0
        return max(self.steps)

    @property
    def steps(self) -> List[int]:
        """ Gets the sorted steps of the timer manager (ie, the union of the steps of all its timers). """
        return sorted(set(step for timer in self.timers.values() for step in timer.steps))

    def get_timer_names(self, timer: Union[None, str, List[str]] = None) -> List[str]:
        """
        Returns the list of existing timer names corresponding to given name(s). Accepts names with wildcards ('*'). If
        no timer name is matched, sends a warning.

        :param timer: if None (default), returns the names of all timers. Otherwise, returns the names of all timers
            matching the given pattern or at least one of the given patterns
        :return: the list of matched timer names
        """
        matches = []
        all_names = list(self.timers.keys())
        if timer is None:
            return all_names
        if isinstance(timer, str):
            timer = [timer]
        for name in all_names:
            if any(compare_string_pattern(name, pattern) for pattern in timer):
                matches.append(name)
        if not matches:
            YAECS_LOGGER.warning(f"WARNING : No existing timer for patterns '{timer}'.")
        return matches

    def render(self, which_step: Union[int, str] = "last", which_timer: Union[None, str, List[str]] = None,
               verbose: Optional[int] = None) -> str:
        """
        Renders a string to display some properties of the timers.

        :param which_step: which timing to get from the timers. Can be an int (index of the step), or 'current'
            (currently running timers), 'last' (last recorded step), 'first' (first recorded step),
            'average' (average of all recorded steps) or 'total' (sum of all recorded steps)
        :param which_timer: which timers to render. Can be None (by default, means all of them), or a timer name or list
            thereof which may contain wildcards ('*')
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail. If None, uses each timer's verbose
        :return: the rendered string
        """
        rendered_string = ""
        which_step = self._process_which(which_step)
        which_timer = self.get_timer_names(timer=which_timer)
        durations = {name: duration for name, duration in self[which_step].items() if name in which_timer}
        verbose = self.verbose if verbose is None else verbose

        # Set header
        if which_step == "current":
            which_string = "(currently) "
        elif which_step == "average":
            which_string = f"(averaged over {self.last_step - self.first_step} steps) "
        elif which_step == "total":
            which_string = f"(accumulated over {self.last_step - self.first_step} steps) "
        else:
            which_string = f"(at step {which_step}) "
        if verbose == 2:
            rendered_string += f"\n---- Reporting on {len(self.timers)} timers {which_string}----\n"
            header_size = len(rendered_string)
        if verbose == 1:
            rendered_string += f"\n{len(self.timers)} timers {which_string}: \n"
        if verbose == 0:
            rendered_string += "{"

        # Set body
        if not durations:
            if verbose in [1, 2]:
                rendered_string += "No timer set so far.\n"
            if verbose == 0:
                rendered_string += " - "
        for name, duration in durations.items():
            if duration is None:
                continue
            if verbose in [1, 2]:
                rendered_string += f" - {name} : {format_duration(duration)}\n"
            if verbose == 0:
                rendered_string += f"{name}:{format_duration(duration)} ; "
                if name == list(durations.keys())[-1]:
                    rendered_string = rendered_string[:-3]

        # Set footer
        if verbose in [1, 2]:
            not_reported = [name for name, duration in durations.items() if duration is None]
            if len(not_reported) > 0:
                reported_string = "running" if which_step == "current" else "recorded "
                reported_string += "so far" if which_step in ["total", "average"] else ""
                reported_string += f"at step {which_step}" if isinstance(which_step, int) else ""
                rendered_string += f"{len(not_reported)} timers not {reported_string} ({', '.join(not_reported)}).\n"
        if verbose == 2:
            rendered_string += "-" * header_size
        if verbose == 0:
            rendered_string += "}"

        return rendered_string

    def reset(self) -> None:
        """ Resets the manager, deleting all timers. """
        self.timers = {}

    def start(self, name: Union[None, str, List[str]] = "MyTimer", step: Optional[int] = None,
              start_time: Optional[float] = None, verbose: Optional[int] = None) -> None:
        """
        Starts a timer.

        :param name: name or list of names of the timers to start, accepts wildcards ('*'). If None, starts all existing
            timers. For names containing a wildcard, starts all existing matching timers.
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param start_time: optional time to start the timer at (if None, uses current time)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        current_time = time.time() if start_time is None else start_time
        if isinstance(name, str):
            name = [name]
        to_start = [] if name is None else [n for n in name if "*" not in n]
        if name is None or any("*" in n for n in name):
            to_start += [n for n in self.get_timer_names(name) if n not in to_start]
        for timer_name in to_start:
            try:
                with TemporaryVerbose(self.timers[timer_name], verbose):
                    self.timers[timer_name].start(step=step, start_time=current_time)
            except KeyError:
                self.timers[timer_name] = Timer(name=timer_name, start=True, step=step, start_time=current_time,
                                                verbose=verbose)

    def stop(self, name: Union[None, str, List[str]] = "MyTimer", step: Optional[int] = None,
             stop_time: Optional[float] = None, verbose: Optional[int] = None) -> Union[None, float, List[float]]:
        """
        Stops a timer.

        :param name: name of the timer or list of timers to stop, accepts wildcards ('*'). None to stop all timers.
        :param step: starting step (if None, assumes step=start step, timings are averages over the elapsed steps)
        :param stop_time: optional time to stop the timer at (if None, uses current time)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :return: the duration of the timer if it was running, otherwise None
        """
        current_time = time.time() if stop_time is None else stop_time
        names = self.get_timer_names(name)
        to_return = []
        for timer_name in names:
            with TemporaryVerbose(self.timers[timer_name], verbose):
                to_return.append(self.timers[timer_name].stop(step=step, stop_time=current_time))
        if not to_return:
            return None if (isinstance(name, str) and "*" not in name) else []
        return to_return[0] if (isinstance(name, str) and "*" not in name) else to_return

    def update(self, start: Union[None, str, List[str]] = None, stop: Union[None, str, List[str]] = None,
               step: Optional[int] = None, update_time: Optional[float] = None,
               verbose: Optional[int] = None) -> List[Optional[float]]:
        """
        Automatically starts and stops timers.

        :param start: names of the timers to start, accepts wildcards ('*')
        :param stop: names of the timers to stop, accepts wildcards ('*')
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param update_time: optional time to update the timers at (if None, uses current time)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        :return: the durations of the stopped timers if any
        """
        current_time = time.time() if update_time is None else update_time
        if start is None:
            start = []
        if isinstance(start, str):
            start = [start]
        if stop is None:
            stop = []
        if isinstance(stop, str):
            stop = [stop]
        timings = [self.stop(name=name, step=step, stop_time=current_time, verbose=verbose) for name in stop]
        for name in start:
            self.start(name=name, step=step, start_time=current_time, verbose=verbose)
        return timings

    def _process_which(self, which: Union[int, str]) -> Union[int, str]:
        """
        Processes 'which' arguments.

        :raises ValueError: if which is not an int or in ['current', 'last', 'first', 'average', 'total']
        :raises IndexError: if which is an int and is out of range
        :return: the processed which
        """
        if isinstance(which, int) and which > self.last_step:
            raise IndexError(f"Unknown step {which} : the last recorded timing in on step {self.last_step}.")
        possible_strings = ["current", "last", "first", "average", "total"]
        if not isinstance(which, int) and which not in possible_strings:
            raise ValueError(f"Invalid value for 'which' : {which}. Can be an int or in {possible_strings}.")
        if which == "last":
            which = self.last_step
        elif which == "first":
            which = self.first_step
        return which


class TemporaryVerbose:
    """ Context manager to change the verbosity of a timer. """
    def __init__(self, timer: Timer, verbose: int):
        """
        Creates a context manager to change the verbosity of a timer.

        :param timer: timer to change the verbosity of
        :param verbose: new verbosity level
        """
        self.timer: Timer = timer
        self.verbose: int = verbose
        self.old_verbose: Optional[int] = None

    def __enter__(self) -> None:
        """ Changes the verbosity level of the timer. """
        if self.verbose is not None:
            self.old_verbose = self.timer.verbose
            self.timer.verbose = self.verbose

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Resets the verbosity level of the timer. """
        if self.old_verbose is not None:
            self.timer.verbose = self.old_verbose


class TimeInContext:
    """ This Context starts a Timer in given TimerManager at the start of the context and stops it at the end. """
    def __init__(self, timer_manager: TimerManager, name: str = "MyTimer", step: Optional[int] = None,
                 verbose: Optional[int] = None):
        """
        Creates a context to start a timer in a timer manager.

        :param timer_manager: timer manager to start the timer in
        :param name: name of the timer
        :param step: starting step (if None, assumes step=last stop step, timings are averages over the elapsed steps)
        :param verbose: verbosity level. 0 is minimal, 1 is normal, 2 is high detail
        """
        self.timer_manager: TimerManager = timer_manager
        self.name: str = name
        self.step: Optional[int] = step
        self.verbose: Optional[int] = verbose

    def __enter__(self) -> None:
        """ Starts the timer in the timer manager. """
        self.timer_manager.start(name=self.name, step=self.step, verbose=self.verbose)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Stops the timer in the timer manager. """
        self.timer_manager.stop(name=self.name, step=self.step, verbose=self.verbose)


def format_duration(duration: float, precision: int = 2, human_readable: bool = True) -> str:
    """ Formats a duration in seconds to a string. """
    if not human_readable:
        return f"{duration:.{precision}f}"
    groups = {"w": 604800, "d": 86400, "h": 3600, "m": 60}
    formatted = ""
    for name, seconds in groups.items():
        if duration >= seconds:
            formatted += f"{int(duration // seconds)}{name} "
            duration %= seconds
    return f"{formatted}{duration:.{precision}f}s"



"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from decimal import Context
import functools
import io
import logging
import re
import sys
from collections.abc import Mapping
from enum import Enum
from numbers import Real
from typing import Any, Callable, Dict, List, Optional, Union

YAECS_LOGGER = logging.getLogger(__name__)
ConfigDeclarator = Union[str, dict]
ConfigInput = Union[List[ConfigDeclarator], ConfigDeclarator]
Hooks = Union[Dict[str, List[str]], List[str]]
TypeHint = Union[type, tuple, list, dict, set, int]
VariationDeclarator = Union[List[ConfigDeclarator], Dict[str, ConfigDeclarator]]
YAML_EXPRESSIONS = {
    "null": re.compile(r'''^(?: ~
                    |null|Null|NULL
                    | )$''', re.X),
    "bool": re.compile(r'''^(?:yes|Yes|YES|no|No|NO
                    |true|True|TRUE|false|False|FALSE
                    |on|On|ON|off|Off|OFF)$''', re.X),
    "int": re.compile(r'''^(?:[-+]?0b[0-1_]+
                    |[-+]?0[0-7_]+
                    |[-+]?(?:0|[1-9][0-9_]*)
                    |[-+]?0x[0-9a-fA-F_]+
                    |[-+]?[1-9][0-9_]*(?::[0-5]?[0-9])+)$''', re.X),
    "float": re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+][0-9]+)?
                    |\.[0-9][0-9_]*(?:[eE][-+][0-9]+)?
                    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
                    |[-+]?\.(?:inf|Inf|INF)
                    |\.(?:nan|NaN|NAN))$''', re.X)
}


class NoValue:
    """ Used to represent a default value not modified by the user. """


class Priority(Enum):
    """ Define priority levels which can be used to qualify when a processing function should be performed. """
    ALWAYS_FIRST = -20
    OFTEN_FIRST = -10
    INDIFFERENT = 0
    SITUATIONAL = 0
    OFTEN_LAST = 10
    ALWAYS_LAST = 20

    def __hash__(self):
        return hash(self.value)

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        if isinstance(other, Real):
            return self.value > other
        if isinstance(other, str):
            return self.value > getattr(self.__class__, other)
        return NotImplemented

    def __rgt__(self, other):
        return self < other

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        if isinstance(other, Real):
            return self.value < other
        if isinstance(other, str):
            return self.value < getattr(self.__class__, other)
        return NotImplemented

    def __rlt__(self, other):
        return self > other

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        if isinstance(other, Real):
            return self.value >= other
        if isinstance(other, str):
            return self.value >= getattr(self.__class__, other)
        return NotImplemented

    def __rge__(self, other):
        return self <= other

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        if isinstance(other, Real):
            return self.value <= other
        if isinstance(other, str):
            return self.value <= getattr(self.__class__, other)
        return NotImplemented

    def __rle__(self, other):
        return self >= other

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value
        if isinstance(other, Real):
            return self.value == other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __req__(self, other):
        return self == other


class TqdmLogFormatter:
    """
    Context setting formatters used in logging handlers for tqdm bars. See https://github.com/tqdm/tqdm/issues/313
    """

    def __init__(self, logger):
        self._logger = logger
        self.__original_formatters = None

    def __enter__(self):
        self.__original_formatters = list()

        for handler in self._logger.handlers:
            self.__original_formatters.append(handler.formatter)

            handler.terminator = ''
            formatter = logging.Formatter('%(message)s')
            handler.setFormatter(formatter)

        return self._logger

    def __exit__(self, exc_type, exc_value, exc_traceback):
        for handler, formatter in zip(self._logger.handlers, self.__original_formatters):
            handler.terminator = '\n'
            handler.setFormatter(formatter)


class TqdmLogger(io.StringIO):
    """File to use in tqdm to make it log its bars to a logger. See https://github.com/tqdm/tqdm/issues/313"""

    def __init__(self, logger):
        super().__init__()

        self._logger = logger

    def write(self, buffer):
        with TqdmLogFormatter(self._logger) as logger:
            logger.info(buffer)

    def flush(self):
        pass


def assign_order(order: Union[Real, 'Priority'] = 0) -> Callable[[Callable], Callable]:
    """
    Decorator used to give an order to a processing function. If several processing functions would be called at a given
    step, they are called in increasing order.

    :param order: order to give the function
    :return: decorated function
    """
    def decorator_order(func: Callable) -> Callable:
        set_function_attribute(func, "order", order)
        return func

    return decorator_order


def assign_yaml_tag(processor_tag: str, processor_type: str,
                    replacement_type_hint: str = "Any") -> Callable[[Callable], Callable]:
    """
    Decorator used to mark a function as a processor added automatically as pre or post processing function (as
    defined by processor_type) to parameters tagged with !type:<processor_tag>. Their type hint will be replaced by
    the type hint defined as replacement_type_hint.

    :param processor_tag: tag to use to mark a param in YAML as auto-processed by this function
    :param processor_type: 'pre' or 'post', type of processing function to add
    :param replacement_type_hint: type hint to use for any param tagged with this auto-processor
    :return: decorated function
    """
    def decorator_tag_assignment(func: Callable) -> Callable:
        func.__dict__["assigned_yaml_tag"] = (processor_tag, processor_type, replacement_type_hint)
        return func

    return decorator_tag_assignment


def compare_string_pattern(name: str, pattern: str) -> bool:
    """
    Returns True when string 'name' matches string 'pattern', with the '*' character matching any number of characters.

    :param name: name to compare
    :param pattern: pattern to match
    :return: result of comparison
    """
    pattern = pattern.strip(" ").split("*")
    if len(pattern) == 1:
        return pattern[0] == name
    if not (name.startswith(pattern[0]) and name.endswith(pattern[-1])):
        return False
    for fragment in pattern:
        index = name.find(fragment)
        if index == -1:
            return False
        name = name[index + len(fragment):]
    return True


def compose(*functions: Callable) -> Callable:
    """
    Returns the composition of the functions given as argument. Functions are applied from left to right, ie :
    compose(f, g, h)(x) = h(g(f(x))).

    :param functions: all functions to compose, applied from left to right
    :return: the composed function
    """
    def compose_2(function_1, function_2):
        def composed(*args, **kwargs):
            return function_2(function_1(*args, **kwargs))
        orders = [get_order(function_1, default=None), get_order(function_2, default=None)]
        orders = [order for order in orders if order is not None]
        if orders:
            set_function_attribute(composed, "order", max(orders))
        hooks = []
        for func in [function_1, function_2]:
            if hasattr(func, "__name__") and func.__name__.startswith("yaecs_config_hook__"):
                hooks += func.__name__.split("__")[1].split(",")
        if hooks:
            set_function_attribute(composed, "__name__", f"yaecs_config_hook__{','.join(list(set(hooks)))}__composed")
        return composed
    return functools.reduce(compose_2, functions, lambda x: x)


def dict_apply(dictionary: dict, function: Callable) -> dict:
    """
    Returns a copy of dict 'dictionary' where function 'function'
    was applied to all values.

    :param dictionary: dictionary to copy
    :param function: function to map
    :return: copied dictionary
    """
    return {k: function(v) for k, v in dictionary.items()}


def escape_symbols(string_to_escape: str, symbols: Union[List[str], str]) -> str:
    """
    Take a string 'string_to_escape' as input and escapes characters
    as defined in 'symbols'.

    :param string_to_escape: string where the escaping operation takes
        place
    :param symbols: list of strings to escape or string containing
        the characters to escape
    :return: escaped string
    """
    for symbol in symbols:
        string_to_escape = string_to_escape.replace(symbol, f"\\{symbol}")
    return string_to_escape


def format_str(config_path_or_dictionary: ConfigDeclarator, size: int = 200) -> str:
    """
    Format helper to shorten configs to display depending on logging level.

    :param config_path_or_dictionary: config to display
    :param size: number of characters allowed to display
    :return: the formatted string
    """
    to_return = str(config_path_or_dictionary)
    if YAECS_LOGGER.getEffectiveLevel() >= logging.INFO:
        return to_return if len(to_return) < size else f"{to_return[:size//2 - 3]} [...] {to_return[-size//2 - 3:]}"
    return to_return


def get_config_from_argv(pattern: str, fallback: Optional[ConfigInput] = None) -> List[str]:
    """
    Get a configuration from the command line arguments.

    :param pattern: pattern to detect in sys.argv
    :param fallback: fallback value if pattern is not detected in sys.argv
    :return: the configuration
    """
    if pattern not in sys.argv and fallback is None:
        raise TypeError(f"The pattern '{pattern}' was not detected in sys.argv.")
    if pattern in sys.argv:
        fallback = [cfg.strip(" ") for cfg in sys.argv[sys.argv.index(pattern) + 1].strip("[]").split(",")]
    if not isinstance(fallback, list):
        fallback = [fallback]
    return fallback


def get_quasi_bash_sys_argv(string_to_convert: str) -> List[str]:
    """
    If a string is passed as input, process it as sys.argv would in a bash shell
    It gives exactly what sys.argv would if the script was used in a bash terminal, except that escaped '!' in quotes
    are properly escaped and the escape symbol is removed, contrary to bash (which would keep the escape for some
    obscure reason).

    :param string_to_convert: string to process
    :return: the list of strings that sys.argv would give
    """
    converted_list = [""]
    in_quotes = ""
    escaped = False
    for index, char in enumerate(string_to_convert):
        if char == "\\" and not escaped and (not in_quotes or string_to_convert[index+1] == "!"):
            escaped = True
        elif char in ['"', "'"] and not escaped:
            if not in_quotes:
                in_quotes = char
            elif in_quotes == char:
                in_quotes = ""
            else:
                converted_list[-1] += char
        elif char == " " and not in_quotes and converted_list[-1] and not escaped:
            converted_list.append("")
        elif char == "!" and not escaped:
            raise ValueError("Bash would say 'event not found', please escape the '!' character.")
        else:
            escaped = False
            converted_list[-1] += char
    if in_quotes:
        raise ValueError(f"Could not parse args : open quotations were left unclosed : {in_quotes}.")
    return converted_list


def get_order(func: Callable, default: Union[None, Real, 'Priority'] = Priority.INDIFFERENT
              ) -> Union[None, Real, 'Priority']:
    """
    If input function has an "order" attribute, returns it. Otherwise, returns the specified "default" value.

    :param func: function to get the order of
    :param default: default value to return if no order is found
    :return: the order value
    """
    return getattr(func, "order", default)


def get_param_as_parsable_string(param: Any) -> str:
    """
    Gets given value as a string that can be parsed by the Configuration. The string is formatted so as to be either
    used as is in a bash shell (ie., python main.py --param_name string), or with merge_from_command_line (ie.,
    config.merge_from_command_line(f"--param_name {string}")

    :param param: parameter value to be returned as a valid string
    :raises TypeError: if the type of 'param' cannot be enforced
    :return: string usable in the command line to reproduce the value of param
    """
    container_separator = ",\\ "
    if param is None:
        return "null"
    if isinstance(param, list):
        parsable_strings = [get_param_as_parsable_string(i) for i in param]
        return f"[{container_separator.join(parsable_strings)}]"
    if isinstance(param, dict):
        parsable_strings = [f"{key}:\\ {get_param_as_parsable_string(value)}" for key, value in param.items()]
        return "{" + container_separator.join(parsable_strings) + "}"
    if isinstance(param, (int, float)) and not isinstance(param, bool):
        return format(Context(prec=20).create_decimal(repr(param)), 'f')
    if isinstance(param, str):
        string = escape_symbols(param, ['"', "'", "!", " "])
        return escape_symbols(f'"{string}"', ['"'])
    if isinstance(param, bool):
        return str(param).lower()
    raise TypeError("Provided value's type is not YAML-compatible (None, str, bool, int, float, list and dict work).")


def hook(hook_name: str) -> Callable[[Callable], Callable]:
    """
    Decorator used to keep track of registered params.

    :param hook_name: name of the hook to store
    :return: decorated function
    """
    def decorator_hook(func: Callable) -> Callable:
        if func.__name__.startswith("yaecs_config_hook__"):
            hooks = func.__name__.split("__")[1].split(",")
            if hook_name in hooks:
                hook_name_in_func_name = ",".join(hooks)
            else:
                hook_name_in_func_name = ",".join(hooks + [hook_name])
            original_name = "__".join(func.__name__.split("__")[2:])
        else:
            hook_name_in_func_name = hook_name
            original_name = func.__name__
        set_function_attribute(func, "__name__", f"yaecs_config_hook__{hook_name_in_func_name}__{original_name}")

        @functools.wraps(func)
        def wrapper_hook(self, *args, **kwargs):
            value = func(self, *args, **kwargs)
            self.add_currently_processed_param_as_hook(hook_name=hook_name)
            return value

        for function_attribute in ["order", "assigned_yaml_tag"]:
            if hasattr(func, function_attribute):
                set_function_attribute(wrapper_hook, function_attribute, getattr(func, function_attribute))
        return wrapper_hook
    for invalid_pattern in ["__", ","]:
        if invalid_pattern in hook_name:
            raise RuntimeError(f"Invalid hook name {hook_name} : '{invalid_pattern}' is not allowed in hook names.")
    return decorator_hook


def is_type_valid(value: Any, config_class: type) -> bool:
    """
    Checks whether input 'value' can be saved in a YAML file by Configuration's YAML Dumper.

    :param value: value to check the type of
    :param config_class: Configuration class, which must be passed as argument to avoid circular imports :(
    :return: result of the test
    """
    if isinstance(value, list):
        return all(is_type_valid(i, config_class) for i in value)
    if isinstance(value, (Mapping, config_class)):
        return all(is_type_valid(i, config_class) for i in value.values())
    return isinstance(value, (int, float, str)) or value is None


def parse_type(string_to_process: str) -> TypeHint:
    """
    Parses an input string containing the type info for a parameter into a complex type as understood by the
    Configuration.check_type function.

    :param string_to_process: string to parse for type
    :return: complex type
    """
    if not string_to_process:
        raise ValueError("Invalid type hint : empty type hint.")
    string = string_to_process.lower()
    mapping_starts = {"tuple_0": "(", "tuple_1": "union[", "nonetuple": "optional[",
                      "list_0": "[", "list_1": "list[",
                      "set_0": "d", "set_1": "dict["}
    types = {"none": None, "int": int, "float": float, "bool": bool, "str": str, "list": list, "dict": dict, "any": 0}
    mapping_ends = {"tuple_0": ")", "tuple_1": "]", "nonetuple": "]",
                    "list_0": "]", "list_1": "]",
                    "set_0": "/d", "set_1": "]"}
    to_return = ("root", [])
    current = []
    current_types = []
    i = 0

    def _get_sub_list(lists, path):
        list_to_get = lists
        for element in path:
            list_to_get = list_to_get[1][element]
        return list_to_get[1]

    def _increment(lists, path, value_to_add, value_type):
        list_to_incr = _get_sub_list(lists, path)
        list_to_incr.append((value_type, value_to_add))

    def _enter_list(lists, path, path_types, path_type):
        list_to_enter = _get_sub_list(lists, path)
        path.append(len(list_to_enter)-1)
        path_types.append(path_type)

    while i < len(string):
        to_find = True
        # Try to detect starts of mappings
        for type_name, fragment in mapping_starts.items():
            if to_find and string[i:i+len(fragment)] == fragment:
                if not (fragment == "d" and string[i:i+len("dict")] == "dict"):
                    to_find = False
                    _increment(to_return, current, [], type_name)
                    _enter_list(to_return, current, current_types, type_name)
                    i += len(fragment)
        # Try to detect simple types
        for fragment, type_name in types.items():
            if to_find and string[i:i+len(fragment)] == fragment:
                to_find = False
                _increment(to_return, current, type_name, "type")
                i += len(fragment)
        # Try to detect commas
        if to_find and string[i] == ",":
            to_find = False
            i += 1
        # Try to detect ends of mappings
        for type_name, fragment in mapping_ends.items():
            if to_find and string[i:i+len(fragment)] == fragment and current_types[-1] == type_name:
                to_find = False
                current = current[:-1]
                current_types = current_types[:-1]
                i += len(fragment)
        if to_find:
            raise ValueError(f"Unexpected token at position {i} : {string_to_process}")

    if current:
        raise ValueError(f"Parsing error : unclosed brackets : {string_to_process}")

    def _struc_to_type(structured_list):
        list_to_consider = structured_list[1]
        if len(list_to_consider) != 1:
            raise ValueError("Parsing error : a source type must contain exactly 1 type (simple or complex) : "
                             f"{string_to_process}")
        if list_to_consider[0][0].startswith("type"):
            return list_to_consider[0][1]
        if list_to_consider[0][0].startswith("tuple"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty tuples are not allowed : {string_to_process}")
            return tuple(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("nonetuple"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty tuples are not allowed : {string_to_process}")
            return (None,) + tuple(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("list"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty lists are not allowed : {string_to_process}")
            return list(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("set"):
            return {"type": _struc_to_type(("", list_to_consider[0][1]))}
        return None

    return _struc_to_type(to_return)


def recursive_set_attribute(obj: Any, key: str, value: Any) -> None:
    """
    Recursively gets attributes of 'obj' until object.__setattr__
    can be used to force-set parameter 'key' to value 'value'.

    :param obj: object where to set the key to the value
    :param key: attribute of the object to set recursively
    :param value: value to set
    """
    if "." in key:
        subconfig, key = key.split(".", 1)
        recursive_set_attribute(obj[subconfig], key, value)
    else:
        object.__setattr__(obj, key, value)


def set_function_attribute(func: Callable, attribute_name: str, value: Any) -> None:
    """
    Adds an attribute to a function or method object.

    :param func: function to add the attribute to
    :param attribute_name: name of the attribute to add
    :param value: value of the attribute
    """
    try:
        setattr(func, attribute_name, value)
    except AttributeError:  # used if func is a method, to modify the underlying function
        setattr(func.__func__, attribute_name, value)


def update_state(state_descriptor: str) -> Callable[[Callable], Callable]:
    """
    Decorator used to store useful information in Configuration._state when using some recursive functions. Kind of a
    hack, but very useful to keep track of the loading state and also to debug.

    :param state_descriptor: string indicating what to store in Configuration._state
    :return: decorated function
    """

    def decorator_update_state(func: Callable) -> Callable:

        @functools.wraps(func)
        def wrapper_update_state(self, *args, **kwargs):
            # State name:
            state_to_append = state_descriptor.split(";")[0]
            for i in state_descriptor.split(";")[1:]:
                # Additional information:
                state_to_append += f";{getattr(self, i)}"
            first_arg = (args[0] if args else (kwargs[list(kwargs.keys())[0]] if kwargs else None))
            self._state.append(  # pylint: disable=protected-access
                state_to_append + f";arg0={first_arg}")  # first arg of function call
            value = func(self, *args, **kwargs)
            self._state.pop(-1)  # pylint: disable=protected-access
            return value

        return wrapper_update_state

    return decorator_update_state



"""
Reactive Reality Machine Learning Config System - ConfigSettersMixin object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from ..yaecs_utils import TypeHint

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigSettersMixin:
    """ Setters Mixin class for YAECS configurations. """

    _main_config: 'Configuration'
    _pre_postprocessing_values: Dict[str, Any]
    _sub_configs_list: List['Configuration']
    _type_hints: Dict[str, TypeHint]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_processing_function(self, param_name: str, function_to_add: Union[str, Callable], processing_type: str
                                ) -> None:
        """ If given parameter does not already have a post-processing function with the same name, adds given function
        as a post-processing function to parameters with the given name. """
        attribute = f"_{processing_type}_processing_functions"
        current_processing = object.__getattribute__(self, attribute)
        set_name = param_name
        while set_name in current_processing:
            set_name = set_name + " "
        new_processing = {set_name: function_to_add, **current_processing}
        object.__setattr__(self, attribute, new_processing)

    def add_processing_function_all(self, param_name: str, function_to_add: Union[str, Callable], processing_type: str
                                    ) -> None:
        """
        Triggers add_processing_function on the main config and all defined and future sub-configs.

        :param param_name: parameter(s) to which to add a postprocessing function. Expects paths with respect to the
            main config.
        :param function_to_add: postprocessing function to add, using the generic name "function" if it has no name
        :param processing_type: choose between 'pre' to add a pre-processing function or 'post' to add a post-processing
            function
        """
        if isinstance(function_to_add, str):
            check_function = self._assigned_as_yaml_tags[function_to_add[len("_tagged_method_"):]][0]
        else:
            check_function = function_to_add
        if hasattr(check_function, "assigned_yaml_tag"):
            if check_function.assigned_yaml_tag[1] != processing_type:
                name = "unknown_function" if not hasattr(check_function, "__name__") else check_function.__name__
                YAECS_LOGGER.warning(f"WARNING : processing function {name} is recommended to use "
                                     f"as {check_function.assigned_yaml_tag[1]}-processing function, "
                                     f"but was declared as {processing_type}-processing function.")
        current_added_processing_name = f"_added_{processing_type}_processing"
        current_added_processing = getattr(self, current_added_processing_name)()
        current_processing_name = f"_{processing_type}_processing_functions"
        current_processing = getattr(self, current_processing_name)
        set_name = param_name
        while set_name in current_processing or set_name in current_added_processing:
            set_name = set_name + " "
        if not any(v == function_to_add if isinstance(function_to_add, str) else v.__name__ == function_to_add.__name__
                   for k, v in current_added_processing.items()
                   if set_name.strip(" ") == k.strip(" ")):
            # Add to main config
            self._main_config.add_processing_function(param_name, function_to_add, processing_type)
            # Add to current sub-configs
            for subconfig in self.get_all_sub_configs():
                subconfig.add_processing_function(param_name, function_to_add, processing_type)
            # Add to future sub-configs
            new_processing = {set_name: function_to_add, **current_added_processing}
            setattr(self.__class__, current_added_processing_name, lambda self: new_processing)

    def add_type_hint(self, name: str, type_hint: TypeHint) -> None:
        """
        Adds a type hint for a parameter to the list of type hints for automatic type checks.

        :param name: full path of the param in the main config
        :param type_hint: type of the param
        """
        self._type_hints[name] = type_hint

    def set_sub_config(self, sub_config: 'Configuration') -> None:
        """
        Registers a new sub-config to the main config.

        :param sub_config: sub-config to register
        """
        if self._are_same_sub_configs(self, self._main_config):
            if all(not self._are_same_sub_configs(i, sub_config) for i in self.get_all_sub_configs()):
                self._sub_configs_list.append(sub_config)
        else:
            self._main_config.set_sub_config(sub_config)

    def remove_value_before_postprocessing(self, name: str) -> None:
        """
        Function used for bookkeeping : it remove a parameter from the pre-post-processing archive.

        :param name: name of the parameter using the dot convention
        """
        if name in self._pre_postprocessing_values:
            del self._pre_postprocessing_values[name]

    def remove_type_hint(self, param_name: str) -> None:
        """
        Removes a registered type hint from a param with given name.

        :param param_name: param from which to remove the type hint
        """
        if param_name in self._type_hints:
            del self._type_hints[param_name]

    def save_value_before_postprocessing(self, name: str, value: Any) -> None:
        """
        Function used for bookkeeping : it saves the value a parameter had before its post-processing.

        :param name: name of the parameter using the dot convention
        :param value: value of the parameter before post-processing
        """
        if name not in self._pre_postprocessing_values:
            self._pre_postprocessing_values[name] = value

    def set_post_processing(self, value: bool = True) -> None:
        """
        Sets the state of the master switch for pre-processing across the entire config object. Calling this for a
        sub-config will also affect the main config and all other sub-configs.

        :param value: value to set the pre-processing to
        """
        object.__setattr__(self._main_config, "_post_process_master_switch", value)

    def set_pre_processing(self, value: bool = True) -> None:
        """
        Sets the state of the master switch for pre-processing across the entire config object. Calling this for a
        sub-config will also affect the main config and all other sub-configs.

        :param value: value to set the pre-processing to
        """
        object.__setattr__(self._main_config, "_pre_process_master_switch", value)

    def set_variation_name(self, value: Optional[str]) -> None:
        """
        Sets the variation name across the entire config object. Calling this for a sub-config will also affect the main
        config and all other sub-configs.

        :param value: value of the new variation name
        """
        object.__setattr__(self._main_config, "_variation_name", value)
        for subconfig in self._main_config.get_all_linked_sub_configs():
            object.__setattr__(subconfig, "_variation_name", value)

    def unset_sub_config(self, sub_config: 'Configuration') -> None:
        """
        Registers a new sub-config to the main config.

        :param sub_config: sub-config to register
        """
        if self._are_same_sub_configs(self, self._main_config):
            if not self._are_same_sub_configs(
                    self._main_config.get(".".join(sub_config.get_nesting_hierarchy()), None), sub_config):
                self._sub_configs_list = [c for c in self.get_all_sub_configs()
                                          if not self._are_same_sub_configs(c, sub_config)]
        else:
            self._main_config.unset_sub_config(sub_config)



""" This module defines the AggregateLogger class. """
import os
import logging
from typing import Any, Dict, List, Optional, Union

from .base_logger import Logger
from .logger_basic import BasicLogger
from .logger_clearml import ClearMLLogger
from .logger_mlflow import MLFlowLogger
from .logger_sacred import SacredLogger
from .logger_tensorboard import TensorBoardLogger

LOGGERS = {
    "basic": BasicLogger,
    "clearml": ClearMLLogger,
    "mlflow": MLFlowLogger,
    "sacred": SacredLogger,
    "tensorboard": TensorBoardLogger,
}
YAECS_LOGGER = logging.getLogger(__name__)


class AggregateLogger:
    """ This class defines an aggregation of different loggers and functions that use aggregated loggers. """
    def __init__(self, tracker, logger_list):
        self.tracker = tracker
        if any(name not in LOGGERS for name in logger_list):
            absent = [name for name in logger_list if name not in LOGGERS]
            raise ValueError(f"Unknown logger(s) {absent}. Available loggers are {list(LOGGERS.keys())}.")
        # Do not track if in a pytorch-lightning spawned process.
        self.types: List[str] = [] if os.getenv('NODE_RANK') else logger_list
        self.logger_list: Dict[Logger] = {name: LOGGERS[name](tracker) for name in logger_list}
        self.logged_artifacts = {}

    def __getitem__(self, item: str) -> Any:
        try:
            return self.logger_list[item].get_logger_object()
        except KeyError:
            return None

    def check_config_requirements(self) -> None:
        """ This checks the config requirements of all loggers. """
        errors = {}
        for name, logger in self.logger_list.items():
            error = logger.check_config_requirements()
            if error != "":
                errors[name] = error
        if errors:
            formatted_errors = "\n".join([f"    {name.upper()}: {error}" for name, error in errors.items()])
            raise RuntimeError(f"Config Requirement Error : errors were met in the following loggers: {errors.keys()}. "
                               f"Find more details below:\n{formatted_errors}")

    def check_install(self) -> None:
        """ This checks the installation of all loggers. """
        errors = {}
        for name, logger in self.logger_list.items():
            error = logger.check_install()
            if error != "":
                errors[name] = error
        if errors:
            formatted_errors = "\n".join([f"    {name}: {error}" for name, error in errors.items()])
            raise RuntimeError(f"Installation Error : errors were met in the following loggers: {errors.keys()}. "
                               f"Find more details below:\n{formatted_errors}")

    @property
    def has_loggers(self) -> bool:
        """ Returns whether the aggregate logger has loggers. """
        return bool(self.logger_list)

    def main_function_context(self) -> list:
        """ Returns a list of context managers that should be used around the main function of the experiment. """
        context_list = []
        for logger in self.logger_list.values():
            context = logger.main_function_context()
            if isinstance(context, list):
                context_list.extend(context)
            else:
                context_list.append(context)
        return context_list

    def modify_main_function(self, main_function):
        """ Returns a modified version of the main function that should be used instead. """
        for logger in self.logger_list.values():
            main_function = logger.modify_main_function(main_function)
        return main_function

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        """ Prepares the logger for the start of the run. """
        for logger in self.logger_list.values():
            logger.start_run(experiment_name, run_name, description, params)

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None,
                   only_loggers: Union[None, str, List[str]] = None,
                   except_loggers: Union[None, str, List[str]] = None) -> None:
        """ Logs a scalar value using the logger. """
        for logger in self._get_loggers(only_loggers, except_loggers):
            logger.log_scalar(name, value, step, sub_logger, description)

    def log_image(self, name: str, image, step: Optional[int] = None, sub_logger: Optional[str] = None,
                  extension: str = "png", only_loggers: Union[None, str, List[str]] = None,
                  except_loggers: Union[None, str, List[str]] = None,
                  maximum: Optional[int] = None, maximum_per_step: Optional[int] = None) -> None:
        """ Logs an image using the logger. The image could be a path to a saved image, matplotlib or plotly figure, a
        PIL.Image, or a n*n*3 numpy array. """
        if self._check_maximums("logged_images", step, maximum, maximum_per_step):
            for logger in self._get_loggers(only_loggers, except_loggers):
                logger.log_image(name, image, step, sub_logger, extension)
            self._add_logged_artifact("logged_images", step)
        else:
            YAECS_LOGGER.debug(f"Image {name} at step {step} was not logged to sub_logger {sub_logger} because the "
                               f"maximum number of images was reached.")

    def _add_logged_artifact(self, artifact_name, step):
        """ Adds an artifact to the logged artifacts. """
        if artifact_name not in self.logged_artifacts:
            self.logged_artifacts[artifact_name] = {step: 1}
        else:
            if step not in self.logged_artifacts[artifact_name]:
                self.logged_artifacts[artifact_name][step] = 1
            else:
                self.logged_artifacts[artifact_name][step] += 1

    def _check_maximums(self, artifact_name, step, maximum, maximum_per_step):
        """ Checks if the maximums have been reached for the given artifact. """
        number = sum(self.logged_artifacts[artifact_name].values()) if artifact_name in self.logged_artifacts else 0
        number_at_step = (0 if number == 0 or step not in self.logged_artifacts[artifact_name]
                          else self.logged_artifacts[artifact_name][step])
        if maximum is not None and number >= maximum >= 0:
            return False
        if maximum_per_step is not None and number_at_step >= maximum_per_step >= 0:
            return False
        return True

    def _get_loggers(self, only_loggers: Union[None, str, List[str]],
                     except_loggers: Union[None, str, List[str]]) -> List[Logger]:
        """ Returns the loggers that should be used for the current function. """
        if only_loggers is None:
            only_loggers = self.logger_list.keys()
        elif isinstance(only_loggers, str):
            only_loggers = [only_loggers]
        only_loggers = [self.logger_list[name] for name in only_loggers if name in self.logger_list]
        if except_loggers is None:
            except_loggers = []
        elif isinstance(except_loggers, str):
            except_loggers = [except_loggers]

        return [logger for logger in only_loggers if not any(logger.name == name for name in except_loggers)]



""" This file defines the base class for all loggers, as well as the LoggerList class to compose loggers. """
import logging
from typing import Any, Optional, Union

YAECS_LOGGER = logging.getLogger(__name__)


class Logger:
    """ Base class for all loggers. Loggers are interfaces to utilities like ClearML, Tensorboard, etc., accessed
    by Experiment and Tracker objects communicate with those utilities. """
    def __init__(self, name: str, tracker):
        self.name = name
        self.tracker = tracker
        self.argument_warnings = set()

    def check_config_requirements(self) -> str:
        """ This returns an empty string if the tracker config and general config satisfy the logger's requirements,
        otherwise the error message if the requirements are not satisfied. """
        return ""

    def check_install(self) -> str:
        """ This returns an empty string if the logger is installed properly, otherwise the error message if the logger
        is not installed properly. """
        return ""

    def get_logger_object(self) -> Any:
        """ Returns the main object repsonsible for the logging activities. """
        return None

    def main_function_context(self):
        """ Returns a context manager or list that should be used around the main function of the experiment. """
        return NoContext(tracker=self.tracker)

    def modify_main_function(self, main_function):
        """ Returns a modified version of the main function that should be used instead. """
        return main_function

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        """ Prepares the logger for the start of the run. """
        raise NotImplementedError

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        """ Logs a scalar value using the logger. """
        raise NotImplementedError

    def log_image(self, name: str, image: Any, step: Optional[int] = None, sub_logger: Optional[str] = None,
                  extension: str = "png") -> None:
        """ Logs an image using the logger. The image could be a path to a saved image, matplotlib or plotly figure, a
        PIL.Image, or a n*n*3 numpy array. """
        raise NotImplementedError

    def _warn_function_argument(self, function_name: str, argument_name: str, argument_value: Any, default_value: Any):
        """ If the user changed the default value, warns the user that a certain argument is not supported by given
        logger, then stores it to avoid repeating the warning. """
        if argument_value != default_value:
            index = function_name + argument_name
            if index not in self.argument_warnings:
                YAECS_LOGGER.warning(f"WARNING : in {function_name} : '{argument_name}' is not used in Logger "
                                     f"{self.name}.")
                self.argument_warnings.add(index)


class NoContext:
    """ A context manager that does nothing. """
    def __init__(self, tracker):
        self.tracker = tracker

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass



"""
Reactive Reality Machine Learning Config System - ConfigHistory object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import glob
import io
import logging
import os
import sys
import time
from pathlib import Path

try:
    import pygraphviz as pgv
except ImportError:
    pgv = None

from .config.config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigHistory:
    """Configuration history manager."""

    def __init__(self, folder_path, config_filters=None,
                 difference_processor=None, ignore_for_graphing=None,
                 add_relevant_edges=False, tolerance=0, group_by=None,
                 metrics=None, display_span=False, config_class=Configuration):
        if pgv is None:
            raise ImportError("ConfigHistory requires pygraphviz - "
                              "currently not installed!")
        self.difference_processor = self.make_processor(difference_processor)
        self.ignore_for_graphing = self.make_processor(ignore_for_graphing)
        self.group_by = group_by

        YAECS_LOGGER.info("Loading configs...")
        if not isinstance(folder_path, dict):
            yaml_files = glob.glob(os.path.join(folder_path, "**/*.yaml"),
                                   recursive=True)
            self.paths = [
                file[:-len("_hierarchy.yaml")] + ".yaml" for file in yaml_files
                if file.endswith("_hierarchy.yaml")
            ]
            self.names = [
                self.get_experiment_name_from_file(file, folder_path)
                for file in self.paths
            ]
            self.folders = [("", folder_path)] * len(self.paths)
        else:
            self.paths = []
            self.names = []
            self.folders = []
            for i in folder_path:
                yaml_files = glob.glob(
                    os.path.join(folder_path[i], "**/*.yaml"), recursive=True)
                to_add = [
                    fil[:-len("_hierarchy.yaml")] + ".yaml"
                    for fil in yaml_files if fil.endswith("_hierarchy.yaml")
                ]
                self.paths += to_add
                self.folders += [(i, folder_path[i])] * len(to_add)
                self.names += [
                    self.get_experiment_name_from_file(fil, folder_path[i],
                                                       name=i)
                    for fil in to_add
                ]

        if config_filters is not None:
            if not isinstance(config_filters, list):
                config_filters = [config_filters]
            for filt in config_filters:
                good = [filt(i) for i in self.paths]
                self.paths = [
                    self.paths[i] for i in range(len(self.paths)) if good[i]
                ]
                self.names = [
                    self.names[i] for i in range(len(self.names)) if good[i]
                ]
                self.folders = [
                    self.folders[i] for i in range(len(self.folders))
                    if good[i]
                ]

        self.modification_times = [os.path.getmtime(fil) for fil in self.paths]
        self.configs = []
        for path in self.paths:
            try:
                sys.stdout = io.StringIO()
                self.configs.append(config_class.load_config(path))
                sys.stdout = sys.__stdout__
            except Exception:
                sys.stdout = sys.__stdout__
                YAECS_LOGGER.error(f"Error while loading config {path}.")
                raise
        for i, config in enumerate(self.configs):
            if config.get("_former_saving_time", None) is not None:
                modification_time = config._former_saving_time
                self.modification_times[i] = modification_time

        YAECS_LOGGER.info(f"Successfully loaded {len(self.configs)} configs from folder(s) {folder_path}. "
                          f"Analysing differences...")

        self.matrix = self.compute_difference_matrix()
        self.span = self.compute_span()
        self.similarity_matrix = [[
            len(
                self.ignore_for_graphing(list(self.matrix[row][col]), row, col,
                                         self))
            for col in range(len(self.matrix[row]))
        ] for row in range(len(self.matrix))]
        self.similarity_coefficients = [
            sum(row) for row in self.similarity_matrix
        ]
        self.metrics = self.compute_metrics(metrics)
        self.config_graph = pgv.AGraph(
            strict=True, directed=True,
            label=self.format_span() if display_span else "")
        self.compute_graph(add_relevant_edges=add_relevant_edges,
                           tolerance=tolerance)
        YAECS_LOGGER.info("Config graph done !")

    def compute_span(self):
        """ Computes span of parameter in experiments. """
        span = {}
        for i, row in enumerate(self.matrix):
            for j, elem in enumerate(row):
                ignored = self.ignore_for_graphing(elem, i, j, self)
                for param in ignored:
                    if param[0] not in span:
                        span[param[0]] = [param[1]]
                    elif param[1] not in span[param[0]]:
                        span[param[0]].append(param[1])
        return span

    def compute_metrics(self, metrics):
        """ Compute all defined metrics to prepare for printing """
        if metrics is None:
            return {}
        if (isinstance(metrics, (list, tuple)) and all(
                isinstance(i, (list, tuple)) and len(i) in [2, 3]
                for i in metrics)):
            pass
        elif isinstance(metrics, (list, tuple)) and len(metrics) in [2, 3]:
            metrics = [metrics]
        else:
            raise TypeError(f"Unrecognised format for metric {metrics}.")
        to_return = {}
        folder_paths = [Path(i).parents[0] for i in self.paths]
        for metric in metrics:
            try:
                metric_value = metric[1](folder_paths)
            except Exception:
                YAECS_LOGGER.error("There was an error parsing metrics.")
                raise
            if len(metric) == 2:
                to_return[metric[0]] = (metric_value, "")
            else:
                to_return[metric[0]] = (metric_value, metric[2])
        return to_return

    def compute_graph(self, add_relevant_edges=False, tolerance=0):
        """ Computes the graph itself before filling it in. """

        def add_directed_edge(first, second):
            if (self.modification_times[first] >
                    self.modification_times[second]):
                self.config_graph.add_edge(
                    second, first, labeljust="l",
                    label=self.format_list(self.matrix[second][first]))
            else:
                self.config_graph.add_edge(
                    first, second, labeljust="l",
                    label=self.format_list(self.matrix[first][second]))

        potential_nodes = list(range(len(self.configs)))
        potential_nodes.sort(key=lambda x: self.modification_times[x])
        potential_nodes.sort(key=lambda x: self.similarity_coefficients[x])
        nodes_added = []
        for _ in range(len(potential_nodes)):
            if not nodes_added:
                index = potential_nodes.pop(0)
                nodes_added.append(index)
                self.config_graph.add_node(
                    index, style="filled", label=(
                        f"{self.names[index]}\n"
                        f"Date : {time.ctime(self.modification_times[index])}"
                        f"{self.format_metrics(index)}"))
            else:
                # Find which node to add in priority
                nodes_by_connectivity = nodes_added.copy()
                nodes_by_connectivity.sort(
                    key=lambda x: -self.modification_times[x])
                nodes_by_connectivity.sort(
                    key=lambda x: self.similarity_coefficients[x])

                relevant_similarities = [[
                    (self.similarity_matrix[i][j] if j in potential_nodes else
                     max(self.similarity_matrix[i]) + 1)
                    for j in range(len(self.similarity_matrix[i]))
                ] for i in nodes_by_connectivity]
                best_new_node = [min(i) for i in relevant_similarities]
                new_parent_node = nodes_by_connectivity[best_new_node.index(
                    min(best_new_node))]
                potential_new_nodes = [
                    i for i in potential_nodes
                    if (self.similarity_matrix[new_parent_node][i] == min(
                        best_new_node))
                ]
                potential_new_nodes.sort(
                    key=lambda x: self.modification_times[x])
                potential_new_nodes.sort(
                    key=lambda x: self.similarity_coefficients[x])
                new_node = potential_nodes.pop(
                    potential_nodes.index(potential_new_nodes[0]))
                # Add the node
                nodes_added.append(new_node)
                date = time.ctime(self.modification_times[new_node])
                self.config_graph.add_node(
                    new_node, style="filled",
                    label=(f"{self.names[new_node]}\n"
                           f"Date : {date}"
                           f"{self.format_metrics(new_node)}"))
                add_directed_edge(new_parent_node, new_node)

        if self.group_by is not None:
            if isinstance(self.group_by, str):
                self.group_by = [self.group_by]
            groups = {}
            for i, config in enumerate(self.configs):
                group = ""
                for j in self.group_by:
                    params = config.get_parameter_names()
                    for param in params:
                        if param.endswith("." + j) or param == j:
                            group += (f" ; {j}:{config[param]}"
                                      if group else f"{j}:{config[param]}")
                if group and group not in groups:
                    groups[group] = [i]
                elif group:
                    groups[group].append(i)
            for group, subgraph in groups.items():
                self.config_graph.add_subgraph(subgraph,
                                               name="cluster_" + group,
                                               label="\n" + group + "\n")

        if add_relevant_edges:
            relevant_similarities = [[
                len(self.matrix[i][j]) if j != i else max(
                    len(self.matrix[i][k])
                    for k in range(len(self.matrix[i]))) + 1
                for j in range(len(self.similarity_matrix[i]))
            ] for i in range(len(self.similarity_matrix))]
            for i, similarities in enumerate(relevant_similarities):
                minimum = min(similarities)
                for j, sim in enumerate(similarities):
                    if sim - minimum < tolerance + 1:
                        if not ((i, j) in self.config_graph.edges() or
                                (i, j) in self.config_graph.edges()):
                            add_directed_edge(i, j)

    # pylint: disable=too-many-locals,too-many-branches
    def compute_colors(self, scheme="date", fill="top", legend=True,
                       base_color="white", class_scheme="/set312/",
                       number_scheme="/blues9/"):
        """ Gets the colour scheme and interpolates it properly. """
        to_remove = []

        # Remove previous legend
        for i in self.config_graph.nodes():
            if "legend" in str(i):
                to_remove.append(i)
        for i in to_remove:
            self.config_graph.remove_node(i)

        # Determine colouring scheme
        unavailable = []
        if scheme == "date":
            indexes_values = list(
                zip(list(range(len(self.configs))), self.modification_times))
            indexes_values.sort(key=lambda x: x[1])
            color_scheme = number_scheme
        elif scheme.startswith("metric:"):
            if scheme[7:] not in self.metrics:
                raise ValueError(f"Unknown metric : {scheme[7:]}.")
            indexes_values = list(
                zip(list(range(len(self.configs))),
                    self.metrics[scheme[7:]][0]))
            for indexes_value in indexes_values:
                if indexes_value[1] is None:
                    unavailable.append((indexes_value[0], None))
            indexes_values = [
                index for index in indexes_values
                if index[0] not in [i[0] for i in unavailable]
            ]
            if all(
                    isinstance(x[1], (int,
                                      float)) and not isinstance(x[1], bool)
                    for x in indexes_values):
                indexes_values.sort(key=lambda x: x[1])
                color_scheme = number_scheme
            else:
                color_scheme = class_scheme
        elif scheme.startswith("param:"):
            values = []
            for config in self.configs:
                found = False
                params = config.get_parameter_names()
                for param in params:
                    if param.endswith("." + scheme[6:]) or param == scheme[6:]:
                        if not found:
                            values.append(config[param])
                            found = True
                        else:
                            raise RuntimeError(f"Ambiguous param : {scheme[6:]}.")
                if not found:
                    raise ValueError(f"Unknown param : {scheme[6:]}.")
            indexes_values = list(zip(list(range(len(self.configs))), values))
            for indexes_value in indexes_values:
                if indexes_value[1] is None:
                    unavailable.append((indexes_value[0], None))
            indexes_values = [
                index for index in indexes_values
                if index[0] not in [i[0] for i in unavailable]
            ]
            if all(
                    isinstance(x[1], (int,
                                      float)) and not isinstance(x[1], bool)
                    for x in indexes_values):
                indexes_values.sort(key=lambda x: x[1])
                color_scheme = number_scheme
            else:
                color_scheme = class_scheme
        else:
            raise ValueError(f"Unrecognised coloring scheme : {scheme}.")
        value_set = list({x[1] for x in indexes_values})
        if color_scheme == number_scheme:
            value_set.sort()
            color_set = [5]
            if fill == "top":
                for _ in range(len(value_set) - 1):
                    color_set = [max([0, min(color_set) - 1])] + color_set
            elif fill == "bottom":
                for _ in range(len(value_set) - 1):
                    color_set += [max([0, min(color_set) - 1])]
            elif fill == "full":
                factor = 6. / float(len(value_set))
                acc = 6 - factor
                for _ in range(len(value_set) - 1):
                    color_set = [max([0, int(acc // 1)])] + color_set
                    acc -= factor
            else:
                raise ValueError(f"Unrecognised fill value : {fill}. Please use 'top', 'bottom' or 'full'.")
        else:
            color_set = [x % 13 for x in range(len(value_set))]
        one_is_colored = False
        for i in self.config_graph.nodes():
            if "legend" not in str(i):
                value = None
                for index_value in indexes_values:
                    if index_value[0] == int(str(i)):
                        value = index_value[1]
                if value is not None:
                    color = color_set[value_set.index(value)]
                    is_none = False
                    for index_value in unavailable:
                        if index_value[0] == int(str(i)):
                            is_none = True
                            i.attr["fillcolor"] = "gray91"
                    if not is_none:
                        one_is_colored = True
                        i.attr["fillcolor"] = (color_scheme + str(color)
                                               if color else base_color)
                else:
                    i.attr["fillcolor"] = "gray91"
        if legend and one_is_colored:
            legend_nodes = []
            for i in range(6 if color_scheme ==
                           number_scheme else len(value_set)):
                if color_scheme == number_scheme:
                    values = [
                        value_set[v] for v in range(len(value_set))
                        if color_set[v] == i
                    ]
                    if not values:
                        label = None
                    elif len(values) == 1:
                        if isinstance(values[0], float) and scheme != 'date':
                            label = f"{values[0]:.3f}"
                        else:
                            label = str(values[0] if scheme != 'date' else time
                                        .ctime(values[0]))
                    else:
                        if isinstance(values[0], float) and scheme != 'date':
                            label = f"[{min(values):.3f} ; {max(values):.3f}]"
                        else:
                            label = ("[" + str(
                                min(values) if scheme != 'date' else time
                                .ctime(min(values))) + " ; " + str(
                                    max(values) if scheme != 'date' else time
                                    .ctime(max(values))) + "]")
                else:
                    if i not in color_set:
                        label = None
                    else:
                        label = ""
                        for j, val in enumerate(value_set):
                            label += f"{val}\n" if color_set[j] == i else ""
                if label is not None:
                    legend_nodes.append("legend_" + str(i))
                    self.config_graph.add_node(
                        "legend_" + str(i), label=label, style="filled",
                        fillcolor=color_scheme + str(i) if i else base_color)
            self.config_graph.add_subgraph(legend_nodes, name="cluster_legend",
                                           label=f"Legend for {scheme}")
            for i in legend_nodes:
                self.config_graph.add_edge(
                    self.modification_times.index(max(
                        self.modification_times)), i, style="invis")

    def draw_graph(self, path="graph.png", scheme="date", fill="top",
                   legend=True):
        """ Function called to draw the computed graph. """
        self.compute_colors(scheme=scheme, fill=fill, legend=legend)
        self.config_graph.layout(prog="dot")
        self.config_graph.draw(path)
        YAECS_LOGGER.info("Config graph saved !")

    def format_metrics(self, index):
        """ Format computed metrics for printing. """
        string = ""
        for i, metric in self.metrics.items():
            if metric[0][index] is None:
                string += f"\n{i} : UNAVAILABLE"
            elif isinstance(metric[0][index], float):
                string += f"\n{i} : {metric[0][index]:.3f} {metric[1]}"
            else:
                string += f"\n{i} : {metric[0][index]} {metric[1]}"
        return string

    def format_span(self):
        """ Format the span of a param over experiments for printing. """
        string = "\nExploration span\n\n"
        for i, val in self.span.items():
            string += f"{i} : {val}" + r"\l"
        return string

    def compute_difference_matrix(self):
        """ Compute difference matrix to prepare for printing. """
        matrix = []
        similars = []
        for i, config_i in enumerate(self.configs):
            row = []
            similar = [i]
            for j, config_j in enumerate(self.configs):
                if j != i:
                    differences = config_i.compare(config_j, reduce=True)
                    row.append(
                        self.difference_processor(differences, i, j, self))
                    if not row[-1]:
                        similar.append(j)
                else:
                    row.append([])
            if all(similar[0] not in k for k in similars):
                similars.append(similar)
            matrix.append(row)
        fuse = []
        for i in similars:
            if len(i) > 1:
                dates = [
                    self.modification_times[j] if j in i else 0
                    for j in range(len(self.modification_times))
                ]
                sort = sorted(i, key=lambda x: dates[x], reverse=True)  # pylint: disable=cell-var-from-loop
                for j in sort:
                    if j != dates.index(max(dates)):
                        self.names[dates.index(max(dates))] = (
                            f"({self.names[j]})\n"
                            + self.names[dates.index(max(dates))])
                fuse += [j for j in i if j != dates.index(max(dates))]
        self.names = [
            self.names[i] for i in range(len(self.names)) if i not in fuse
        ]
        self.configs = [
            self.configs[i] for i in range(len(self.configs)) if i not in fuse
        ]
        self.paths = [
            self.paths[i] for i in range(len(self.paths)) if i not in fuse
        ]
        self.modification_times = [
            self.modification_times[i]
            for i in range(len(self.modification_times)) if i not in fuse
        ]
        matrix = [[
            matrix[i][j] for j in range(len(matrix[i])) if j not in fuse
        ] for i in range(len(matrix)) if i not in fuse]
        return matrix

    @staticmethod
    def get_experiment_name_from_file(file, folder, name=None):
        """ Extracts the experiment name from the name of a file saved in an experiment folder. """
        file_parent_minus_folder = os.path.relpath(
            Path(file).parents[0], os.path.commonpath([folder,
                                                       file])).strip("/_")
        if name is None:
            return file_parent_minus_folder
        return f"{name}:{file_parent_minus_folder}"

    @staticmethod
    def format_list(list_to_format):
        """ Formats list for printing """
        string = " "
        for i in list_to_format:
            string += str(i[0]) + " : " + str(i[1]) + r"\l "
        return string

    @staticmethod
    def make_processor(argument):
        """ Returns a processor created from input. """

        def apply_list(list_to_apply, inp, *args):
            for func in list_to_apply:
                inp = func(inp, *args)
            return inp

        if argument is None:
            return lambda d, i, j, h: d
        if not isinstance(argument, list):
            return argument
        return lambda d, i, j, h: apply_list(argument, d, i, j, h)



"""
Reactive Reality Machine Learning Config System - Usage example
(project- specific Configuration sub-class)
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import random

from yaecs import Configuration


def check_model_type(model_type):
    assert model_type in ["linear", "polynomial"]
    return model_type


def check_params(param):
    if param is None:
        return None
    if param == "random":
        return random.random()
    assert isinstance(param, (int, float))
    return param


# This is literally the copy/pasted template given in
# framework/config/default_config_template
# There is however a few pre-processing functions added to provide with
# further guidance as to how to modify this class
class ProjectSpecificConfiguration(Configuration):
    """Default configuration template."""

    @staticmethod
    def get_default_config_path():
        return "./configs/default/main_config.yaml"

    def parameters_pre_processing(self):
        return {"model.type": check_model_type,
                "model.param*": check_params,
                "*_variation": self.register_as_config_variations,
                "grid": self.register_as_grid,
                "*path_to_config": self.register_as_additional_config_file,
                "*paths_to_configs": lambda x: [
                    self.register_as_additional_config_file(path)
                    for path in x]}



""" This file defines the class for the basic logger. """
import logging
import os
from shutil import copyfile
import sys
import traceback
from typing import Any, Optional, Union
import warnings

from mock import patch

from .base_logger import Logger
from .logger_utils import NotImportedModule, add_to_csv, new_print

try:
    import matplotlib
except ImportError:
    matplotlib = NotImportedModule("matplotlib")
try:
    import numpy as np
except ImportError:
    np = NotImportedModule("numpy")
try:
    from PIL import Image
except ImportError:
    Image = NotImportedModule("PIL")
try:
    import plotly
except ImportError:
    plotly = NotImportedModule("plotly")

YAECS_LOGGER = logging.getLogger(__name__)


class BasicLogger(Logger):
    """ Basic Logger. This logger logs everything needed, albeit in a very simple way. """
    def __init__(self, tracker):
        super().__init__("Basic Logger", tracker)
        self.path = None

    def check_config_requirements(self) -> str:
        try:
            self.path = self.tracker.experiment.config.get_experiment_path()
        except RuntimeError:
            self.path = self.tracker.config.get("basic_logdir", None)
        if self.path is None:
            error = ("Config Requirement Error : no path was provided for the basic logger. Provide one using an "
                     "experiment path in the config or by setting the basic_logdir key in the tracker config.")
            return error
        return ""

    def get_logger_object(self) -> Any:
        return self.path

    def main_function_context(self):
        basic_tracker_context = BasicTrackerContext(self.path,
                                                    self.tracker.experiment.number_of_runs,
                                                    self.tracker.experiment.current_run)
        return [basic_tracker_context, patch('builtins.print', side_effect=new_print)]

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        self.tracker.experiment.config.save(os.path.join(self.path, "config.yaml"))
        with open(os.path.join(self.path, "comments.txt"), 'w') as file:
            file.write(description)

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        self._warn_function_argument("log_scalar", "description", description, None)
        extended_name = name
        if sub_logger is not None:
            extended_name = f"{sub_logger}/{extended_name}"
        add_to_csv(os.path.join(self.path, "logged_scalars.csv"),
                   extended_name, value, -1 if step is None else step)

    def log_image(self, name: str, image: Any, step: Optional[int] = None, sub_logger: Optional[str] = None,
                  extension: str = "png") -> None:
        # Paths and names
        directory = [self.path, "images"]
        if sub_logger:
            directory.append(sub_logger)
        if os.path.dirname(name):
            directory.append(os.path.dirname(name))
        if step is not None:
            directory.append(f"step_{step}")
        os.makedirs(image_path := os.path.join(*directory), exist_ok=True)
        output_path = os.path.join(image_path, f"{os.path.basename(name)}.{extension}")

        # Logging
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            if isinstance(image, str):
                if not os.path.isfile(image):
                    raise FileNotFoundError(f"Cannot log image '{image}' : path does not exist !")
                copyfile(image, output_path[:-len(extension)] + image.split(".")[-1])
            elif isinstance(image, np.ndarray):
                Image.fromarray(image.astype(np.uint8)).save(output_path)
            elif isinstance(image, Image.Image):
                image.save(output_path)
            elif isinstance(image, matplotlib.figure.Figure):
                image.savefig(output_path)
            elif isinstance(image, plotly.graph_objs.Figure):
                plotly.io.write_image(image, output_path)
            else:
                YAECS_LOGGER.warning(f"WARNING : image {name} at step {step} was not logged to sub_logger {sub_logger}"
                                     f" : unrecognised type {type(image)}.")


class BasicTrackerContext:
    """ Class used to set up the context for YAECS' basic tracker """

    def __init__(self, logger_path: str, runs: Optional[int], current: Optional[int]):
        """
        Initialises a context used to declare the loggers required by the basic tracker.

        :param logger_path: path used by the basic tracker to log
        :param runs: number of runs in the experiment
        :param current: index of current run from 0 to runs-1
        """
        self.runs, self.current = runs, current
        self.logging_handlers = [logging.FileHandler(os.path.join(logger_path, "stdout.log")),
                                 logging.StreamHandler(sys.stdout)]
        self.logging_handlers[0].setFormatter(logging.Formatter(fmt="%(asctime)s : [TRACKER] %(message)s"))
        self.logging_handlers[1].setFormatter(logging.Formatter(fmt="[TRACKER] %(message)s"))
        self.config_handler = logging.FileHandler(os.path.join(logger_path, "stdout.log"))
        self.config_handler.setFormatter(logging.Formatter(fmt="%(asctime)s : [CONFIG] %(message)s"))
        if "pytorch_lightning" in logging.root.manager.loggerDict:
            if not logging.getLogger("pytorch_lightning").propagate:
                self.pytorch_handler = logging.FileHandler(
                    os.path.join(logger_path, "stdout.log"))
                self.pytorch_handler.setFormatter(logging.Formatter(fmt="%(asctime)s : [%(levelname)s] %(message)s"))
        self.print_handlers = [logging.FileHandler(os.path.join(logger_path, "stdout.log")),
                               logging.StreamHandler(sys.stdout)]
        self.print_handlers[0].setFormatter(logging.Formatter(fmt="%(asctime)s : [%(levelname)s] %(message)s"))

    def __enter__(self):
        # Setting up YAECS loggers
        for handler in self.logging_handlers:
            logging.getLogger().addHandler(handler)
        logging.root.setLevel(logging.INFO)
        run_count = ("" if self.runs is None or self.current is None or self.runs == 1
                     else f" {self.current + 1}/{self.runs}")
        logging.root.info(f"Starting experiment{run_count}...\n")
        y_logger = logging.getLogger("yaecs")
        if not y_logger.propagate:
            y_logger.addHandler(self.config_handler)

        # Setting up external modules' compatibility loggers
        if "pytorch_lightning" in logging.root.manager.loggerDict:
            pl_logger = logging.getLogger("pytorch_lightning")
            if not pl_logger.propagate:
                pl_logger.addHandler(self.pytorch_handler)

        # Setting up print catcher
        print_catcher = logging.getLogger("yaecs.print_catcher")
        print_catcher.propagate = False
        for handler in self.print_handlers:
            print_catcher.addHandler(handler)

    def __exit__(self, exc_type, exc_val, exc_tb):
        print_catcher = logging.getLogger("yaecs.print_catcher")

        # Logging error
        if exc_type is not None:
            message_lines = traceback.format_exc().split("\n")
            print_catcher.error("\n".join([message_lines[0]] + message_lines[3:]))

        # Unsetting print catcher
        ([h for h in print_catcher.handlers if isinstance(h, logging.FileHandler)][0]).setFormatter(
            logging.Formatter(fmt="%(message)s"))
        print_catcher.info("")
        logging.root.info("Experiment has concluded !")
        print_catcher.propagate = True
        for handler in self.print_handlers:
            print_catcher.removeHandler(handler)

        # Unsetting YAECS loggers
        for handler in self.logging_handlers:
            logging.getLogger().removeHandler(handler)
        logging.getLogger("yaecs").removeHandler(self.config_handler)

        # Unsetting external modules' compatibility loggers
        if "pytorch_lightning" in logging.root.manager.loggerDict:
            logging.getLogger("pytorch_lightning").removeHandler(self.pytorch_handler)



"""
Reactive Reality Machine Learning Config System - unit tests
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os

import pytest

from yaecs.user_utils import make_config


@pytest.fixture
def tmp_file_name(tmpdir):
    index = 1 + max([-1] + [
        int(fil[3:-5]) for fil in os.listdir(tmpdir)
        if fil.startswith('tmp') and fil.endswith('.yaml')
    ])
    filename = tmpdir / f'tmp{index}.yaml'
    yield filename


@pytest.fixture
def yaml_default(tmpdir):
    index = len(os.listdir(tmpdir))
    content = (f"param1: 0.1\n--- !subconfig1\nparam2: 3.0\n---\n"
               f"def_second_path: '{tmpdir / f'default_second{index}.yaml'}'\n"
               "exp_second_path: null")
    with open(tmpdir / f'default{index}.yaml', "w", encoding='utf-8') as fil:
        fil.write(content)
    content = ("--- !subconfig2\nparam3: 20.0\nsubconfig3:\n  "
               "param4: 'string'")
    with open(tmpdir / f'default_second{index}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'default{index}.yaml')


@pytest.fixture
def yaml_experiment(tmpdir):
    index = len(os.listdir(tmpdir))
    content = ("subconfig1.param2: 2.0\n\nexp_second_path: \""
               f"experiment_second{index}.yaml\"")
    with open(tmpdir / f'experiment{index}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    content = "--- !subconfig2\nparam3: 30.0"
    with open(tmpdir / f'experiment_second{index}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'experiment{index}.yaml')


@pytest.fixture
def yaml_default_unlinked(tmpdir):
    content = "param1: !type:dict\n  param2: 1\n  param3:\n    param4: 2"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_default_preproc_default_dot_param(tmpdir):
    content = "param1.param2: 2"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_default_sub_variations(tmpdir):
    content = "param1:\n  var: []"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_default_set_twice(tmpdir):
    content = "param: 2"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_experiment_sub_star(tmpdir):
    content = "subconfig2:\n  '*param*': 1.0"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_experiment_sub_dot(tmpdir):
    content = "subconfig2:\n  subconfig3.param4: 1.0"
    with open(tmpdir / f'tmp{len(os.listdir(tmpdir))}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'tmp{len(os.listdir(tmpdir))-1}.yaml')


@pytest.fixture
def yaml_craziest_config(tmpdir):
    with open(tmpdir / 'd_first.yaml', "w", encoding='utf-8') as fil:
        fil.write("p1: 1\n"
                  "--- !c1\n"
                  "c2:\n"
                  f"  f_path: {'d_second.yaml'}\n"
                  "---\n"
                  "c4.p3: 3\n"
                  "c4.p7: 7\n"
                  f"f_path: {tmpdir / 'd_second.yaml'}")
    with open(tmpdir / 'd_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("--- !c3\n"
                  "p2: 2\n"
                  f"c5.s_path: {tmpdir / 'd_third.yaml'}\n"
                  "---\n"
                  "p6: 6")
    with open(tmpdir / 'd_third.yaml', "w", encoding='utf-8') as fil:
        fil.write("--- !c6\n"
                  "p4: !type:(dict,int) {a: 4}\n"
                  "---\n"
                  "p5: 5")
    with open(tmpdir / 'e_first.yaml', 'w', encoding='utf-8') as fil:
        fil.write("'*p6': 7\n"
                  "--- !c1\n"
                  f"c2.f_path: {tmpdir / 'e_second.yaml'}\n"
                  "---\n"
                  "c4:\n"
                  "  p3: 'test'\n"
                  "c4.p7: 'test2'")
    with open(tmpdir / 'e_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("--- !c3\n"
                  "'*.p*': 8\n"
                  "'*p4': {b: 5}")
    yield str(tmpdir / 'd_first.yaml'), str(tmpdir / 'e_first.yaml')


@pytest.fixture
def yaml_no_file_call_processing_while_loading(tmpdir):
    with open(tmpdir / 'd_first.yaml', "w", encoding='utf-8') as fil:
        fil.write(f"test_path: {tmpdir / 'd_second.yaml'}\nexp_path: null")
    with open(tmpdir / 'd_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.1")
    with open(tmpdir / 'e_first.yaml', "w", encoding='utf-8') as fil:
        fil.write(f"exp_path: {tmpdir / 'e_second.yaml'}")
    with open(tmpdir / 'e_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.2")
    config = make_config(str(tmpdir / 'd_first.yaml'),
                         str(tmpdir / 'e_first.yaml'),
                         additional_configs_suffix="_path",
                         variations_suffix="var*", grids_suffix="grid",
                         do_not_merge_command_line=True)
    config.save(str(tmpdir / 'save.yaml'))
    with open(tmpdir / 'd_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.3")
    with open(tmpdir / 'e_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.4")
    config2 = make_config(str(tmpdir / 'd_first.yaml'),
                          str(tmpdir / 'save.yaml'),
                          additional_configs_suffix="_path",
                          variations_suffix="var*", grids_suffix="grid",
                          do_not_merge_command_line=True)
    yield config, config2


@pytest.fixture
def yaml_no_file_call_processing_while_loading_nested(tmpdir):
    with open(tmpdir / 'nd_first.yaml', "w", encoding='utf-8') as fil:
        fil.write(f"c:\n  test_path: {tmpdir / 'nd_second.yaml'}\n  "
                  "exp_path: null")
    with open(tmpdir / 'nd_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.1")
    with open(tmpdir / 'ne_first.yaml', "w", encoding='utf-8') as fil:
        fil.write(f"c:\n  exp_path: {tmpdir / 'ne_second.yaml'}")
    with open(tmpdir / 'ne_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.2")
    config = make_config(str(tmpdir / 'nd_first.yaml'),
                         str(tmpdir / 'ne_first.yaml'),
                         additional_configs_suffix="_path",
                         variations_suffix="var*", grids_suffix="grid",
                         do_not_merge_command_line=True)
    config.save(str(tmpdir / 'nsave.yaml'))
    with open(tmpdir / 'nd_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.3")
    with open(tmpdir / 'ne_second.yaml', "w", encoding='utf-8') as fil:
        fil.write("param: 0.4")
    config2 = make_config(str(tmpdir / 'nd_first.yaml'),
                          str(tmpdir / 'nsave.yaml'),
                          additional_configs_suffix="_path",
                          variations_suffix="var*", grids_suffix="grid",
                          do_not_merge_command_line=True)
    yield config, config2


@pytest.fixture
def yaml_type_check(tmpdir):
    index = len(os.listdir(tmpdir))
    content = (f"root_path: '{tmpdir / f'default_second{index}.yaml'}'\nsubconfig:\n  "
               f"sub_path: '{tmpdir / f'default_second{index}.yaml'}'")
    with open(tmpdir / f'default{index}.yaml', "w", encoding='utf-8') as fil:
        fil.write(content)
    content = ("param_int: !type:int 1\nparam_float: !type:float 0.1\nparam_str: !type:str abc\n"
               "param_bool: !type:bool false\n"
               "param_none: !type:none null\nparam_list: !type:list [0]\nparam_dict: !type:dict\n  a: 0\n"
               "param_any: !type:Any null\nparam_tuple1: !type:(str,float) 0.4\nparam_tuple2: !type:(str,float) ab\n"
               "param_optional1: !type:Optional[int] null\nparam_optional2: !type:Optional[int] 3\n"
               "param_listint: !type:List[int] [0]\nparam_listintstr: !type:List[int,str] [0, ab]\n"
               "param_dictlistoptionalint: !type:Dict[List[Optional[int]]]\n  a: [0, 1, null, 3]")
    with open(tmpdir / f'default_second{index}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'default{index}.yaml')


@pytest.fixture
def yaml_tag_assignment_check(tmpdir):
    index = len(os.listdir(tmpdir))
    content = (f"param1: !type:add_1 0.1\n--- !subconfig1\nparam2: 3.0\n---\n"
               f"def_second_path: '{tmpdir / f'default_second{index}.yaml'}'\n"
               "exp_second_path: null")
    with open(tmpdir / f'default{index}.yaml', "w", encoding='utf-8') as fil:
        fil.write(content)
    content = ("--- !subconfig2\nparam3: 20.0\nsubconfig3:\n  "
               "param4: !type:copy 'param1'")
    with open(tmpdir / f'default_second{index}.yaml', "w",
              encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'default{index}.yaml')


@pytest.fixture
def config_vs_dict_checks(tmpdir):
    index = len(os.listdir(tmpdir))
    content = ("config1:\n"
               "  config2:\n"
               "    param1: !type:dict\n"
               "      key1: !type:dict\n"
               "        key2: 0\n"
               "      key3: ['!type:dict']\n"
               "--- !config3.config4\n"
               "config5:\n"
               "  config6:\n"
               "    param1: !type:dict\n"
               "      key1: !type:dict\n"
               "        key1: 0\n"
               "      key2: ['!type:dict']\n"
               "--- !config10.config11\n"
               "config1.config7: !main_2\n"
               "  config8:\n"
               "    param1: !type:dict\n"
               "      key1: !type:dict\n"
               "        key1: 0\n"
               "      key2: ['!type:dict']\n"
               "  config9.param2: null")
    with open(tmpdir / f'default{index}.yaml', "w", encoding='utf-8') as fil:
        fil.write(content)
    yield str(tmpdir / f'default{index}.yaml')



"""
Reactive Reality Machine Learning Config System - ConfigGettersMixin object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

from ..yaecs_utils import TypeHint, get_param_as_parsable_string

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigGettersMixin:
    """ Getters Mixin class for YAECS configurations. """

    __getattribute__: Callable[[str], Any]
    _main_config: 'Configuration'
    _methods: List[str]
    _modified_buffer: List[str]
    _name: str
    _nesting_hierarchy: List[str]
    _operating_creation_or_merging: bool
    _pre_post_processing_values: Dict[str, Any]
    _protected_attributes: List[str]
    _reference_folder: Optional[str]
    _state: List[str]
    _sub_configs_list: List['Configuration']
    _type_hints: Dict[str, TypeHint]
    _variation_name: str
    _was_last_saved_as: Optional[str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, parameter_name: str, default_value: Any) -> Any:
        """
        Behaves similarly to dict.get(parameter_name, default_value)

        :param parameter_name: parameter to query
        :param default_value: value to return if the parameter does not exist
        :return: queried value
        """
        try:
            return self[parameter_name]
        except (AttributeError, TypeError):
            return default_value

    def get_all_linked_sub_configs(self) -> List['Configuration']:
        """
        Returns the list of all sub-configs that are directly linked to the root config by a chain of other sub-configs.
        For this to be the case, all of those sub-configs need to be contained directly in a parameter of another
        sub-config. For example, a sub-config stored in a list that is a parameter of a sub-config is not linked.

        :return: list corresponding to the linked sub-configs
        """
        all_linked_configs = []
        for i in self._get_user_defined_attributes():
            object_to_scan = getattr(self, "___" + i if i in self._methods else i)
            if isinstance(object_to_scan, ConfigGettersMixin):
                all_linked_configs = all_linked_configs + [object_to_scan] + object_to_scan.get_all_linked_sub_configs()
        return all_linked_configs

    def get_all_sub_configs(self) -> List['Configuration']:
        """
        Returns the list of all sub-configs, including sub-configs of other sub-configs

        :return: list corresponding to the sub-configs
        """
        if self._are_same_sub_configs(self, self._main_config):
            return list(self._sub_configs_list)
        return self._main_config.get_all_sub_configs()

    def get_command_line_argument(self, deep: bool = True, do_return_string: bool = False) -> Union[List[str], str]:
        """
        Returns a list of command line parameters that can be used in a bash shell to re-create this exact config
        from the default. Can alternatively return the string itself with do_return_string=True.

        :param deep: whether to also take the sub-config parameters into account
        :param do_return_string: whether to return a string (True) or a list of strings (False, default)
        :return: list or string containing the parameters
        """
        to_return = []
        for param in self.get_parameter_names(deep=deep):
            if not isinstance(self[param], ConfigGettersMixin):
                full_name = self._get_full_path(param)
                value = get_param_as_parsable_string(
                    self[param] if full_name not in self.get_main_config().get_pre_post_processing_values() else
                    self.get_main_config().get_pre_post_processing_values()[full_name],
                )
                to_return.append(f"--{param} {value}")

        return " ".join(to_return) if do_return_string else to_return

    def get_dict(self, deep: bool = True, pre_post_processing_values: bool = False) -> dict:
        """
        Returns a dictionary corresponding to the config.

        :param deep: whether to recursively turn sub-configs into dicts or keep them as sub-configs
        :param pre_post_processing_values: whether to return the values before the post-processing step
        :return: dictionary corresponding to the config
        """
        to_return = {}
        for key in self._get_user_defined_attributes():
            if deep and isinstance(self[key], ConfigGettersMixin):
                to_return[key] = self[key].get_dict(deep=True, pre_post_processing_values=pre_post_processing_values)
            else:
                full_name = self._get_full_path(key)
                to_return[key] = (self.get_main_config().get_pre_post_processing_values().get(full_name, self[key])
                                  if pre_post_processing_values else self[key])
        return to_return

    def get_main_config(self) -> 'Configuration':
        """
        Getter for the main config corresponding to this config or sub-config. Using this is often hacky.

        :return: the main config
        """
        return self._main_config

    def get_master_switch(self, processing_type: str) -> bool:
        """
        Getter for either the pre- or post-processing master switch depending on processing_type

        :param processing_type: can be 'pre' or 'post' : processing type to get the master switch for
        :return: the main config
        """
        return self._pre_process_master_switch if processing_type == "pre" else self._post_process_master_switch

    def get_modified_buffer(self) -> List[str]:
        """
        Getter for the buffer of modified parameters corresponding to this config or sub-config. This gets filled during
        a creation or merging operation to keep track of all the parameters modified by this operation. Then, it is
        emptied as all modified parameters get post-processed before the end of the operation.

        :return: the buffer of modified elements
        """
        return self._modified_buffer

    def get_name(self) -> str:
        """
        Returns the name of the config. It is composed of a specified part (or 'main' when unspecified) and an indicator
        of its index in the list of variations of its parent if it is a variation of a config. This indicator is
        prefixed by '_VARIATION_'.

        :return: string corresponding to the name
        """
        variation_suffix = ("_VARIATION_" + self._variation_name if self._variation_name is not None else "")
        return self._name + variation_suffix

    def get_nesting_hierarchy(self) -> List[str]:
        """
        Returns the nesting hierarchy of the config

        :return: list corresponding to the nesting hierarchy
        """
        return self._nesting_hierarchy

    def get_parameter_names(self, deep: bool = True, no_sub_config: bool = False) -> List[str]:
        """
        Returns the list of the names all parameters in this config. If deep is true, also returns the names of the
        parameters in the sub-configs using the dot convention.

        :param deep: whether to also return the names of the parameters in the sub-configs
        :param no_sub_config: if True, exclude names of sub-configs and only return real parameters
        :return: the list of the names of all parameters
        """
        complete_list = self._get_user_defined_attributes(no_sub_config=no_sub_config)
        if deep:
            order = len(self.get_nesting_hierarchy())
            for subconfig in self.get_all_linked_sub_configs():
                complete_list += [
                    ".".join(subconfig.get_nesting_hierarchy()[order:] + [param])
                    for param in subconfig.get_parameter_names(deep=False, no_sub_config=no_sub_config)
                ]
        return complete_list

    def get_pre_post_processing_values(self) -> Dict[str, Any]:
        """
        Returns a dictionary containing :

        * as keys : all the names of the parameters which have been post-processed
        * as values : the values those parameters had before the post-processing operation

        In particular, those values are the ones used when saving the config.

        :return: dictionary of values before post-processing
        """
        return self._pre_postprocessing_values

    def get_processed_param_name(self, full_path: bool = True) -> str:
        """
        When used in a processing function, returns the full path to that param in the config, or its path in self if
        full_path is False.

        :param full_path: if True, returns the path to the param in the main config, otherwise the path to the param in
            self.
        """
        if full_path:
            return self._get_full_path(self._get_param_name_from_state())
        return self._get_param_name_from_state()

    def get_save_file(self) -> Optional[str]:
        """
        If the config was saved previously, returns the path to this save. Otherwise, returns None.

        :return: the path to the save if it exists, None otherwise
        """
        return self._was_last_saved_as

    def get_reference_folder(self) -> Optional[str]:
        """
        If a reference folder has been registered, returns it. Otherwise, returns None.

        :return: the reference folder if it exists, None otherwise
        """
        return self._reference_folder

    def get_type_hint(self, param_name) -> TypeHint:
        """
        Returns the type hint for this param, or 0 if it has no type hint.

        :return: possible types for this param according to its type hint, 0 if no type hint defined
        """
        if param_name in self._type_hints:
            return self._type_hints[param_name]
        if "." in param_name:
            return self[".".join(param_name.split(".")[:-1])].get_type_hint(param_name.split(".")[-1])
        return 0

    def get_type_hints(self) -> Dict[str, TypeHint]:
        """
        Returns all the type hints for this config

        :return: dictionary with keys being parameter names and values being their possible types
        """
        return self._type_hints

    def get_variation_name(self) -> str:
        """
        Returns the variation name of the config

        :return: variation name
        """
        return self._variation_name

    def is_in_operation(self) -> bool:
        """
        Returns whether the config is currently in a creation or merging process.

        :return: True if the config is in a creation or merging process, False otherwise
        """
        return self._operating_creation_or_merging

    def _get_full_path(self, param_name: str) -> str:
        """ Get the full name of given param in the main config """
        return ".".join(self._nesting_hierarchy + [param_name])

    def _get_param_name_from_state(self) -> str:
        """ If there is a param processing in the state stack, returns the name of the param. """
        name = None
        for state in self._state[::-1]:
            if state.startswith("processing"):
                if state.count(";arg0=") > 1:
                    raise ValueError("How did you even manage to raise this ?")
                name = state.split(";arg0=")[-1]
                break
        if name is None:
            raise RuntimeError("Processing function was called outside a processing phase.")
        return name

    def _get_tagged_methods_info(self) -> List[Tuple[Union[str, Callable]]]:
        """ Returns a list of info on the methods which were assigned a YAML tag. """
        all_methods = [getattr(self, name) for name in self._methods]
        return [getattr(method, "assigned_yaml_tag") + (method,)
                for method in all_methods if hasattr(method, "assigned_yaml_tag")]

    def _get_user_defined_attributes(self, no_sub_config: bool = False) -> List[str]:
        """ Frequently used to get a list of the names of all the parameters that were in the user's config. """
        return [
            i[3:] if i.startswith("___") else i
            for i in self.__dict__
            if (i not in self._protected_attributes + ["config_metadata"]
                and not (no_sub_config and isinstance(self[i], ConfigGettersMixin)))
        ]



import os
from shutil import rmtree

from .test_project_config.config import UnifiedSegmentationConfig
from .utils import comparison_strings
from yaecs import Experiment


def compare_file_content_to_string(file_path, string, skip_file_first_line=False):
    with open(file_path, "r") as f:
        lines = f.readlines()
    if skip_file_first_line:
        lines = lines[1:]
    assert "".join(lines) == string


def main(config, tracker):
    print(config.details())
    for i in range(10):
        with tracker.measure_time("tracking time"):
            tracker.log_scalar("scalar", i, sub_logger="train")
            tracker.log_scalar("scalar", i+2, sub_logger="test")
        tracker.step()


def test_config_project():
    """ Tests loading a complex config from a sample project. """
    config = UnifiedSegmentationConfig.build_from_argv(
        fallback=["unittests/experiment/test_project_config/experiments/s",
                  "unittests/experiment/test_project_config/experiments/variations"]
    )
    Experiment(config, main).run(run_description="description")
    compare_file_content_to_string(os.path.join(config.b, "f_var_2", "comments.txt"),
                                   comparison_strings["comment"])
    compare_file_content_to_string(os.path.join(config.b, "f_var_2", "config.yaml"),
                                   comparison_strings["config"], True)
    compare_file_content_to_string(os.path.join(config.b, "f_var_2", "config_hierarchy.yaml"),
                                   comparison_strings["hierarchy"])
    # Delete temporary folder
    rmtree("./tmp")



"""
Reactive Reality Machine Learning Config System - Usage example (main file)
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import os

from configs.project_config import ProjectSpecificConfiguration
from project_utils.utils import (create_and_train_model, create_data,
                                 log_experiment, test_model_and_return_metrics)

from yaecs import ConfigHistory


def perform_experiment(config_path):
    config = ProjectSpecificConfiguration.load_config(config_path)

    for _ in range(config.number_of_repeats):
        if config.get_config_variations():
            variations = config.get_config_variations()
        else:
            variations = [config]
        for variation in variations:
            train, test = create_data(variation.data)
            model = create_and_train_model(variation.model, train)
            computed_metrics = test_model_and_return_metrics(
                model, test, variation.metrics)
            log_experiment(variation, computed_metrics)


if __name__ == "__main__":
    if "log" not in os.listdir():
        # Perform random search
        print("\nStep 0 : performing dummy test")
        perform_experiment("./configs/experiment/dummy_test.yaml")

        # Perform random search
        print("\nStep 1 : performing random search")
        perform_experiment("./configs/experiment/random_search.yaml")

        # Perform grid search
        print("\nStep 2 : performing grid search")
        perform_experiment("./configs/experiment/grid_search.yaml")

    # Build config graphs
    # First let's build the graph of all the experiments. We want to filter
    # out the dummy_test, so we'll include a config filter. A config filter is
    # a function which takes a path as argument and returns whether the
    # experiment corresponding to this path should appear in the graph
    # (True/False).

    def config_filter(path):
        return "dummy_test" not in path

    history = ConfigHistory("log", config_filters=config_filter,
                            config_class=ProjectSpecificConfiguration)
    history.draw_graph("all_experiments.png")

    # It's a big mess, isn't it ? Let's improve it step by step :)
    # First, we should remove the parameter names which we
    # don't care about : type_variation, param1_variation,
    # param2_variation, param3_variation, grid and number_of_repeats.
    # This can be done by processing the differences between configs
    # with a difference_processor. A difference_processor is a function
    # which takes the list of all differences, the index of the
    # reference config, the index of the compared config, and the
    # history object itself. It returns the new list of differences.
    # The three last arguments are here to provide information to use
    # during processing. We don't need them in our case. Since the
    # difference list contains ("name", difference) tuples, we just
    # need to access the names for our purpose.

    def difference_processor(difference, reference_index, compared_index,
                             hist):
        # NOTE we don't use all the arguments here but they are required
        # by ConfigHistory
        # pylint: disable=unused-argument
        return [
            d for d in difference if d[0] not in [
                "type_variation", "param1_variation", "param2_variation",
                "param3_variation", "grid", "number_of_repeats"
            ]
        ]

    history = ConfigHistory("log", config_filters=config_filter,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("all_experiments_processed_differences.png")

    # But there's still way too many nodes here, and just removing the
    # dummy run did not help much.
    # Let's have a look at the random search and the grid search
    # separately.

    history = ConfigHistory("log/random_search", config_filters=config_filter,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("random_search_processed_differences.png")
    history = ConfigHistory("log/grid_search", config_filters=config_filter,
                            add_relevant_edges=True,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("grid_search_processed_differences.png")

    # This already looks much easier to parse ! But the colors are
    # useless, because we have performed all those experiments at the
    # same time and the graph is colored by date. We would like to
    # color the graph with respect to our metrics. First, let's implement
    # the metrics. The metrics is a list of ("name", function) tuple, where
    # the function takes the path to an experiment and reads the corresponding
    # metric value from the folder.
    def get_metric(paths, key_name):
        res = []
        for path in paths:
            with open(os.path.join(path, "metrics.json"), "r",
                      encoding="utf-8") as fil:
                res.append(json.load(fil)[key_name])
        return res

    metrics = [
        ("exponential_l2", lambda paths: get_metric(paths, 'is_exponential')),
        ("square_root_l2", lambda paths: get_metric(paths, 'is_square_root'))
    ]

    random_hist = ConfigHistory("log/random_search",
                                config_filters=config_filter, metrics=metrics,
                                config_class=ProjectSpecificConfiguration,
                                difference_processor=difference_processor)
    grid_hist = ConfigHistory("log/grid_search", config_filters=config_filter,
                              metrics=metrics, add_relevant_edges=True,
                              config_class=ProjectSpecificConfiguration,
                              difference_processor=difference_processor)

    # We can now draw the graphs using the metrics as a way to color.
    # By default (fill="top"), the colors are applied only to the configs
    # where the value is the highest. In our case, we want to color all node
    # so we'll use fill="full", where the whitest nodes are the best.

    random_hist.draw_graph("random_search_exp_colour.png",
                           scheme="metric:exponential_l2", fill="full")
    random_hist.draw_graph("random_search_sqrt_colour.png",
                           scheme="metric:square_root_l2", fill="full")
    grid_hist.draw_graph("grid_search_exp_colour.png",
                         scheme="metric:exponential_l2", fill="full")
    grid_hist.draw_graph("grid_search_sqrt_colour.png",
                         scheme="metric:square_root_l2", fill="full")

    # Finally, it can be very interesting to study the effect of a given
    # parameter on the metrics. One way to do this is to group the graph by
    # groups where this parameter takes the same value.

    history = ConfigHistory("log/random_search", config_filters=config_filter,
                            metrics=metrics, group_by=["type"],
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("random_search_exp_by_type.png",
                       scheme="metric:exponential_l2", fill="full")
    history.draw_graph("random_search_sqrt_by_type.png",
                       scheme="metric:square_root_l2", fill="full")

    # On these graphs, it seems that a polynomial model is always better,
    # w.r.t. "exp" or "sqrt". This is why the grid search is performed with
    # a polynomial model.

    history = ConfigHistory("log/grid_search", config_filters=config_filter,
                            metrics=metrics, group_by=["param1"],
                            add_relevant_edges=True,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("grid_search_exp_by_param1.png",
                       scheme="metric:exponential_l2", fill="full")
    history.draw_graph("grid_search_sqrt_by_param1.png",
                       scheme="metric:square_root_l2", fill="full")
    history = ConfigHistory("log/grid_search", config_filters=config_filter,
                            metrics=metrics, group_by=["param2"],
                            add_relevant_edges=True,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("grid_search_exp_by_param2.png",
                       scheme="metric:exponential_l2", fill="full")
    history.draw_graph("grid_search_sqrt_by_param2.png",
                       scheme="metric:square_root_l2", fill="full")
    history = ConfigHistory("log/grid_search", config_filters=config_filter,
                            metrics=metrics, group_by=["param3"],
                            add_relevant_edges=True,
                            config_class=ProjectSpecificConfiguration,
                            difference_processor=difference_processor)
    history.draw_graph("grid_search_exp_by_param3.png",
                       scheme="metric:exponential_l2", fill="full")
    history.draw_graph("grid_search_sqrt_by_param3.png",
                       scheme="metric:square_root_l2", fill="full")

    # Looking at those graphs makes it easy to see that, for each metric,
    # there is a value of the studied parameter which is ALWAYS better.
    # For instance, note how in grid_search_exp_by_param1.png, following an
    # arrow from a node in the param1:0.5 group to its counterpart in the
    # param1:1 group always makes the color go lighter.



""" Experiment utility functions. """
from typing import List, Union


def format_mode(mode: Union[None, str, List[str]]) -> str:
    """ Formats the experiment mode for the purpose of the experiment comment. """
    modes = {
        "TRAINING": ["train", "learn", "fit"],
        "TESTING": ["test", "evaluate", "benchmark"],
        "VALIDATION": ["val"],
        "INFERENCE": ["infer", "predict", "forecast", "apply"],
        "DATA PROCESSING": ["process", "preprocess", "data"],
        "DEBUG": ["debug", "inspect"],
    }
    if mode is None:
        return "None"
    if isinstance(mode, str):
        mode = [mode]
    formatted = []
    for mode_ in mode:
        matched = f'"{mode_.upper()}"'
        for key, values in modes.items():
            if any(mode_.lower().startswith(value) for value in values):
                matched = key
                break
        formatted.append(matched)
    return ", ".join(formatted)



"""
Reactive Reality Machine Learning Config System
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import sys

# Logging configuration code is reused from the __init__.py in pytorch-lightning.
# See : https://github.com/Lightning-AI/lightning/blob/master/src/pytorch_lightning/__init__.py
_ROOT_LOGGER = logging.root
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)

# if root logger has handlers, propagate messages up and let root logger process them
if not _ROOT_LOGGER.hasHandlers():
    _HANDLER = logging.StreamHandler(sys.stdout)
    _HANDLER.setFormatter(logging.Formatter(fmt="[CONFIG] %(message)s"))
    _LOGGER.addHandler(_HANDLER)
    _LOGGER.propagate = False

from .config.config import Configuration  # pylint: disable=wrong-import-position # noqa: E402
from .config_history import ConfigHistory  # pylint: disable=wrong-import-position # noqa: E402
from .experiment.experiment import Experiment  # pylint: disable=wrong-import-position # noqa: E402
from .user_utils import (  # pylint: disable=wrong-import-position # noqa: E402
    get_template_class,
    make_config,
    tqdm_file,
)
from .yaecs_utils import (  # pylint: disable=wrong-import-position # noqa: E402
    assign_order,
    assign_yaml_tag,
    Priority,
)

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown_version")

__all__ = ['__version__', 'assign_order', 'assign_yaml_tag', 'ConfigHistory', 'Configuration', 'Experiment',
           'get_template_class', 'make_config', 'Priority', 'tqdm_file', 'version_tuple']



"""
Reactive Reality Machine Learning Config System - unit tests
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from yaecs.user_utils import get_template_class, make_config


def load_config(*configs, default_config=None, preprocessing=None,
                postprocessing=None):
    return make_config(default_config, *configs,
                       pre_processing_dict=preprocessing,
                       post_processing_dict=postprocessing,
                       additional_configs_suffix="_path",
                       variations_suffix="var*", grids_suffix="grid",
                       do_not_merge_command_line=True)


def template(default_config=None):
    return get_template_class(default_config_path=default_config,
                              additional_configs_suffix="_path", tracker_config="tracker_config",
                              variations_suffix="var*", grids_suffix="grid")



""" This file defines the class for the MLFlow logger. """
import importlib.util
import logging
from typing import Any, Optional, Union

from .base_logger import Logger
from .logger_utils import NotImportedModule, value_to_float

try:  # MLFlow integration, see original package : https://mlflow.org
    import mlflow
except ImportError:
    mlflow = NotImportedModule("mlflow")

YAECS_LOGGER = logging.getLogger(__name__)


class MLFlowLogger(Logger):
    """ MLFlow Logger. This logger logs things to a MLFlow tracking_uri which can be a server or a folder. """
    def __init__(self, tracker):
        super().__init__("MLFlow Logger", tracker)
        self.run = None
        self.tracking_uri = self.tracker.config.get("tracking_uri", None)

    def check_config_requirements(self) -> str:
        if self.tracking_uri is None:
            return ("Config Requirement Error : no tracking_uri was provided for the mlflow logger. Provide one using "
                    "the tracking_uri key in the tracker config.")
        return ""

    def check_install(self) -> str:
        if not importlib.util.find_spec("mlflow"):
            return "Install Error : your experiment tracking config requires mlflow - currently not installed !"
        return ""

    def get_logger_object(self) -> Any:
        return self.run

    def main_function_context(self):
        return MLFContext()

    def start_run(self, experiment_name: str, run_name: str, description: str, params: dict) -> None:
        mlflow.set_tracking_uri(self.tracking_uri)
        mlflow.set_experiment(experiment_name)
        self.run = mlflow.start_run(run_name=run_name, description=description)
        mlflow.log_params(params)

    def log_scalar(self, name: str, value: Union[float, int], step: Optional[int] = None,
                   sub_logger: Optional[str] = None, description: Optional[str] = None) -> None:
        self._warn_function_argument("log_scalar", "description", description, None)
        value = value_to_float(value, self.name)
        if value == "":
            return

        extended_name = name
        if sub_logger is not None:
            extended_name = f"{sub_logger}/{extended_name}"
        mlflow.log_metric(key=extended_name, value=value, step=step)


class MLFContext:
    """ Class used to set up the context for mlflow's tracker """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        mlflow.end_run()



""" Defines some utility functions and objects for the experiment tests. """


comparison_strings = {
    "comment": """[2/2] VALIDATION run in tmp/normal/test_0/f_var_2 (variation f_var_2) :
Purpose : description""",
    "hierarchy": """config_hierarchy:
- /home/vac/Documents/Projects/ait-yaecs/unittests/experiment/test_project_config/defaults/default.yaml
- unittests/experiment/test_project_config/experiments/s
- unittests/experiment/test_project_config/experiments/variations
- a_config:
    ts:
      t4:
        fr: b
        use: true
  mode: validation
""",
    "config": """a: false
b: tmp/normal/test
c: Try something
d: v
e: 1
f:
  var_1:
    a_config:
      ts:
        t4:
          use: false
    mode: training
  var_2:
    a_config:
      ts:
        t4:
          use: true
          fr: b
    mode: validation
mode: validation
a_config:
  a: true
  t:
    s: 1024
    pp:
      p:
      - t1
      - t2
      - t3
      - t5
      - t4
      sp: {}
  v:
    s:
    - 320
    - 480
    pp:
      p:
      - t1
      - t5
      - t4
      sp:
        t1.t1m: false
        t2.t1m: false
  te:
    s:
    - 320
    - 480
    pp:
      p:
      - t1
      - t5
      - t4
      sp:
        t1.t1m: false
        t2.t1m: false
  ts:
    t1:
      use: true
      t1m: true
    t2:
      use: true
      t1m: true
    t3:
      use: false
      t3_probability: 0.4
      m: 0
    t4:
      use: true
      fr: b
    t5:
      use: true
      nm: 1
a_config_file: a_config.yaml
b_config:
  atc: svb
b_config_file: b_config.yaml
t:
  tf: folder
  tn: 50
v:
  vf: 5
c_config_file: c_config.yaml
d_config_file: unittests/experiment/test_project_config/general/bsc
tracker_config:
  type:
  - basic
  - clearml
  project_name: yaecs_dev_tests
""",
}
